import json
import requests

NOTION_API_TOKEN = 'secret_Adp7AI8ugJpicvcKQDAbjpuN7Ah6fbgJ6JnVilOHzFj'
NOTION_DATABASE_ID = "958ba2d81d194c1fa86accf65c1f6b9e"
NOTION_DATABASE_URL = f'https://api.notion.com/v1/databases/{NOTION_DATABASE_ID}/query'


def get_text_from_res(res, key):
    return res[key]['title'][0]['plain_text']


def get_single_from_res(res, key):
    return res[key]['select']['name']

def get_multi_from_res(res, key):
    items = []
    for item in res[key]['multi_select']:
        items.append(item['name'])
    return items


def get_value_from_res(res, key):
    return res[key]['number']


def get_checkbox_from_res(res, key):
    return res[key]['checkbox']


def get_link_from_res(res, key):
    urls = {}
    for item in res[key]['files']:
        name = item['name']
        assert name.replace(".pth", "") == key.lower()
        url = item['file']['url']
        urls[name] = url
    return urls[name]


def print_format(name, key):
    print("{0:<15}: ".format(name) + str(key))

def get_link_by_id(id, flag):
    headers = {'Authorization': f'Bearer {NOTION_API_TOKEN}',
               'Notion-Version': '2022-02-22',
               'Content-Type': 'application/json'}
    payload = {'page_size': 100}
    has_more = True
    while has_more is True:
        res = requests.post(NOTION_DATABASE_URL, data=json.dumps(payload), headers=headers).json()
        has_more = res['has_more']
        payload = {'start_cursor':res['next_cursor']}

        for r in res['results']:
            prop = r['properties']
            if len(prop["ID"]['title']) == 0:
                continue
            if get_text_from_res(prop, "ID") != id:
                continue
            print("="*60)
            print(id)
            print("="*60)
            
            print_format("Method", get_single_from_res(prop, "Method"))
            print_format("Architecture", get_single_from_res(prop, "Architecture"))
            print_format("Batch Size", get_single_from_res(prop, "Batch Size"))
            print_format("Aug.", get_checkbox_from_res(prop, "Aug."))
            print_format("AWP", get_checkbox_from_res(prop, "AWP"))
            print_format("Extra Data", get_checkbox_from_res(prop, "Extra Data"))
            for measure in ["Clean(Last)", "PGD(Last)", "Clean(Best)", "PGD(Best)"]:
                value = get_value_from_res(prop, measure)
                if value:
                    print_format(measure, "%2.2f%%"%value)
#             print("Detail Info: "+r['url'])
            print("="*60)
            return get_link_from_res(prop, flag)



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
from argparse import ArgumentParser, Namespace
from typing import Final

from airtk.defense import (
    AdtppTraining,
    AdtTraining,
    CurratTraining,
    FatTraining,
    FeatureScatterTraining,
    GairatTraining,
    TradesawpTraining,
    TradesTraining,
    VaTraining,
    YopoTraining,
)


def get_args() -> Namespace:
    parser: ArgumentParser = ArgumentParser(
        "AirTk CLI Interface",
        description="Run adversarial training experiments.",
    )
    parser.add_argument(
        "training_method",
        type=str,
        choices=[
            "adt",
            "adtpp",
            "currat",
            "fat",
            "fs",
            "gairat",
            "trades",
            "tradesawp",
            "va",
            "yopo",
        ],
    )
    parser.add_argument("dataset", type=str)
    parser.add_argument("model", type=str)
    parser.add_argument("out_dir", type=str)
    parser.add_argument("--epochs", "-e", type=int, default=100)
    parser.add_argument("--batch-size", "-bs", type=int, default=256)
    parser.add_argument("--learning-rate", "-lr", type=float, default=0.9)
    return parser.parse_args()


def main() -> None:
    args: Final[Namespace] = get_args()
    if args.training_method == "adt":
        AdtTraining(
            args.dataset,
            args.model,
            batch_size=args.batch_size,
            epochs=args.epochs,
            lr=args.learning_rate,
            model_dir=args.out_dir,
        )()
    elif args.training_method == "adtpp":
        AdtppTraining(
            args.dataset,
            args.model,
            batch_size=args.batch_size,
            epochs=args.epochs,
            lr=args.learning_rate,
            model_dir=args.out_dir,
        )()
    elif args.training_method == "currat":
        CurratTraining(
            args.dataset,
            args.model,
            batch_size=args.batch_size,
            epochs=args.epochs,
            lr=args.learning_rate,
            model_dir=args.out_dir,
        )()
    elif args.training_method == "fat":
        FatTraining(
            args.dataset,
            args.model,
            out_dir=args.out_dir,
            epochs=args.epochs,
            lr=args.learning_rate,
        )()
    elif args.training_method == "fs":
        FeatureScatterTraining(
            args.dataset,
            10 if args.dataset in {"cifar10", "mnist"} else 100,
            epochs=args.epochs,
            lr=args.learning_rate,
            model_dir=args.out_dir,
        )()
    elif args.training_method == "gairat":
        GairatTraining(
            args.dataset,
            args.model,
            epochs=args.epochs,
            lr_max=args.learning_rate,
            model_dir=args.out_dir,
        )()
    elif args.training_method == "trades":
        TradesTraining(
            args.dataset,
            args.model_dir,
            model_ir=args.out_dir,
            epochs=args.epochs,
            lr=args.learning_rate,
        )()
    elif args.training_method == "tradesawp":
        TradesawpTraining(
            args.dataset,
            args.model,
            model_idr=args.out_dir,
            epochs=args.epochsm,
            lr=args.learning_rate,
        )()
    elif args.training_method == "va":
        VaTraining(
            args.dataset,
            args.model,
            model_dir=args.out_dir,
            epochs=args.epochs,
            lr=args.learning_rate,
        )()
    elif args.training_method == "yopo":
        YopoTraining(
            args.dataset, model_dir=args.out_dir, epochs=args.epochs
        )()


if __name__ == "__main__":
    main()



# None attacks
from .attacks.apgd import APGD
from .attacks.apgdt import APGDT
from .attacks.autoattack import AutoAttack
from .attacks.bim import BIM

# L2 attacks
from .attacks.cw import CW
from .attacks.deepfool import DeepFool
from .attacks.difgsm import DIFGSM
from .attacks.eaden import EADEN

# L1 attacks
from .attacks.eadl1 import EADL1
from .attacks.eotpgd import EOTPGD

# Linf, L2 attacks
from .attacks.fab import FAB
from .attacks.ffgsm import FFGSM

# Linf attacks
from .attacks.fgsm import FGSM
from .attacks.gn import GN
from .attacks.jitter import Jitter
from .attacks.jsma import JSMA
from .attacks.mifgsm import MIFGSM
from .attacks.nifgsm import NIFGSM
from .attacks.onepixel import OnePixel
from .attacks.pgd import PGD
from .attacks.pgdl2 import PGDL2
from .attacks.pgdrs import PGDRS
from .attacks.pgdrsl2 import PGDRSL2
from .attacks.pifgsm import PIFGSM
from .attacks.pifgsmpp import PIFGSMPP
from .attacks.pixle import Pixle
from .attacks.rfgsm import RFGSM
from .attacks.sinifgsm import SINIFGSM

# L0 attacks
from .attacks.sparsefool import SparseFool
from .attacks.spsa import SPSA
from .attacks.square import Square
from .attacks.tifgsm import TIFGSM
from .attacks.tpgd import TPGD
from .attacks.upgd import UPGD
from .attacks.vanila import VANILA
from .attacks.vmifgsm import VMIFGSM
from .attacks.vnifgsm import VNIFGSM
from .wrappers.lgv import LGV

# Wrapper Class
from .wrappers.multiattack import MultiAttack

__version__ = "3.5.1"
__all__ = [
    "VANILA",
    "GN",
    "FGSM",
    "BIM",
    "RFGSM",
    "PGD",
    "EOTPGD",
    "FFGSM",
    "TPGD",
    "MIFGSM",
    "UPGD",
    "APGD",
    "APGDT",
    "DIFGSM",
    "TIFGSM",
    "Jitter",
    "NIFGSM",
    "PGDRS",
    "SINIFGSM",
    "VMIFGSM",
    "VNIFGSM",
    "SPSA",
    "JSMA",
    "EADL1",
    "EADEN",
    "PIFGSM",
    "PIFGSMPP",
    "CW",
    "PGDL2",
    "DeepFool",
    "PGDRSL2",
    "SparseFool",
    "OnePixel",
    "Pixle",
    "FAB",
    "AutoAttack",
    "Square",
    "MultiAttack",
    "LGV",
]
__wrapper__ = [
    "LGV",
    "MultiAttack",
]



import torch


from airtk.model import WideResNet

if __name__ == "__main__":
    model = WideResNet(num_classes=100)
    model.load_state_dict(torch.load("data/model/cifar100/clean-cifar100-wideres34.pt"))
    print(model.forward(torch.zeros([1, 3, 32, 32])))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import time
import math

import torch
import torch.nn.functional as F

from ..attack import Attack


class Square(Attack):
    r"""
    Square Attack in the paper 'Square Attack: a query-efficient black-box adversarial attack via random search'
    [https://arxiv.org/abs/1912.00049]
    [https://github.com/fra31/auto-attack]

    Distance Measure : Linf, L2

    Arguments:
        model (nn.Module): model to attack.
        norm (str): Lp-norm of the attack. ['Linf', 'L2'] (Default: 'Linf')
        eps (float): maximum perturbation. (Default: 8/255)
        n_queries (int): max number of queries (each restart). (Default: 5000)
        n_restarts (int): number of random restarts. (Default: 1)
        p_init (float): parameter to control size of squares. (Default: 0.8)
        loss (str): loss function optimized ['margin', 'ce'] (Default: 'margin')
        resc_schedule (bool): adapt schedule of p to n_queries (Default: True)
        seed (int): random seed for the starting point. (Default: 0)
        verbose (bool): print progress. (Default: False)
        targeted (bool): targeted. (Default: False)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.Square(model, model, norm='Linf', eps=8/255, n_queries=5000, n_restarts=1, eps=None, p_init=.8, seed=0, verbose=False, targeted=False, loss='margin', resc_schedule=True)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        norm="Linf",
        eps=8 / 255,
        n_queries=5000,
        n_restarts=1,
        p_init=0.8,
        loss="margin",
        resc_schedule=True,
        seed=0,
        verbose=False,
    ):
        super().__init__("Square", model)
        self.norm = norm
        self.n_queries = n_queries
        self.eps = eps
        self.p_init = p_init
        self.n_restarts = n_restarts
        self.seed = seed
        self.verbose = verbose
        self.loss = loss
        self.rescale_schedule = resc_schedule
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)
        adv_images = self.perturb(images, labels)

        return adv_images

    def margin_and_loss(self, x, y):
        """
        :param y:        correct labels if untargeted else target labels
        """

        logits = self.get_logits(x)
        xent = F.cross_entropy(logits, y, reduction="none")
        u = torch.arange(x.shape[0])
        y_corr = logits[u, y].clone()
        logits[u, y] = -float("inf")
        y_others = logits.max(dim=-1)[0]

        if not self.targeted:
            if self.loss == "ce":
                return y_corr - y_others, -1.0 * xent
            elif self.loss == "margin":
                return y_corr - y_others, y_corr - y_others
        else:
            if self.loss == "ce":
                return y_others - y_corr, xent
            elif self.loss == "margin":
                return y_others - y_corr, y_others - y_corr

    def init_hyperparam(self, x):
        assert self.norm in ["Linf", "L2"]
        assert not self.eps is None
        assert self.loss in ["ce", "margin"]

        if self.device is None:
            self.device = x.device
        self.orig_dim = list(x.shape[1:])
        self.ndims = len(self.orig_dim)
        if self.seed is None:
            self.seed = time.time()

    def check_shape(self, x):
        return x if len(x.shape) == (self.ndims + 1) else x.unsqueeze(0)

    def random_choice(self, shape):
        t = 2 * torch.rand(shape).to(self.device) - 1
        return torch.sign(t)

    def random_int(self, low=0, high=1, shape=[1]):
        t = low + (high - low) * torch.rand(shape).to(self.device)
        return t.long()

    def normalize_delta(self, x):
        if self.norm == "Linf":
            t = x.abs().view(x.shape[0], -1).max(1)[0]
            return x / (t.view(-1, *([1] * self.ndims)) + 1e-12)

        elif self.norm == "L2":
            t = (x ** 2).view(x.shape[0], -1).sum(-1).sqrt()
            return x / (t.view(-1, *([1] * self.ndims)) + 1e-12)

    def lp_norm(self, x):
        if self.norm == "L2":
            t = (x ** 2).view(x.shape[0], -1).sum(-1).sqrt()
            return t.view(-1, *([1] * self.ndims))

    def eta_rectangles(self, x, y):
        delta = torch.zeros([x, y]).to(self.device)
        x_c, y_c = x // 2 + 1, y // 2 + 1

        counter2 = [x_c - 1, y_c - 1]
        for counter in range(0, max(x_c, y_c)):
            delta[
                max(counter2[0], 0) : min(counter2[0] + (2 * counter + 1), x),
                max(0, counter2[1]) : min(counter2[1] + (2 * counter + 1), y),
            ] += (
                1.0 / (torch.Tensor([counter + 1]).view(1, 1).to(self.device) ** 2)
            )  # nopep8
            counter2[0] -= 1
            counter2[1] -= 1

        delta /= (delta ** 2).sum(dim=(0, 1), keepdim=True).sqrt()

        return delta

    def eta(self, s):
        delta = torch.zeros([s, s]).to(self.device)
        delta[: s // 2] = self.eta_rectangles(s // 2, s)
        delta[s // 2 :] = -1.0 * self.eta_rectangles(s - s // 2, s)
        delta /= (delta ** 2).sum(dim=(0, 1), keepdim=True).sqrt()
        if torch.rand([1]) > 0.5:
            delta = delta.permute([1, 0])

        return delta

    def p_selection(self, it):
        """ schedule to decrease the parameter p """

        if self.rescale_schedule:
            it = int(it / self.n_queries * 10000)

        if 10 < it <= 50:
            p = self.p_init / 2
        elif 50 < it <= 200:
            p = self.p_init / 4
        elif 200 < it <= 500:
            p = self.p_init / 8
        elif 500 < it <= 1000:
            p = self.p_init / 16
        elif 1000 < it <= 2000:
            p = self.p_init / 32
        elif 2000 < it <= 4000:
            p = self.p_init / 64
        elif 4000 < it <= 6000:
            p = self.p_init / 128
        elif 6000 < it <= 8000:
            p = self.p_init / 256
        elif 8000 < it:
            p = self.p_init / 512
        else:
            p = self.p_init

        return p

    def attack_single_run(self, x, y):
        with torch.no_grad():
            adv = x.clone()
            c, h, w = x.shape[1:]
            n_features = c * h * w
            n_ex_total = x.shape[0]

            if self.norm == "Linf":
                x_best = torch.clamp(
                    x + self.eps * self.random_choice([x.shape[0], c, 1, w]), 0.0, 1.0
                )
                margin_min, loss_min = self.margin_and_loss(x_best, y)
                n_queries = torch.ones(x.shape[0]).to(self.device)
                s_init = int(math.sqrt(self.p_init * n_features / c))

                for i_iter in range(self.n_queries):
                    idx_to_fool = (margin_min > 0.0).nonzero().flatten()

                    if len(idx_to_fool) == 0:
                        break

                    x_curr = self.check_shape(x[idx_to_fool])
                    x_best_curr = self.check_shape(x_best[idx_to_fool])
                    y_curr = y[idx_to_fool]
                    if len(y_curr.shape) == 0:
                        y_curr = y_curr.unsqueeze(0)
                    margin_min_curr = margin_min[idx_to_fool]
                    loss_min_curr = loss_min[idx_to_fool]

                    p = self.p_selection(i_iter)
                    s = max(int(round(math.sqrt(p * n_features / c))), 1)
                    vh = self.random_int(0, h - s)
                    vw = self.random_int(0, w - s)
                    new_deltas = torch.zeros([c, h, w]).to(self.device)
                    new_deltas[:, vh : vh + s, vw : vw + s] = (
                        2.0 * self.eps * self.random_choice([c, 1, 1])
                    )

                    x_new = x_best_curr + new_deltas
                    x_new = torch.min(
                        torch.max(x_new, x_curr - self.eps), x_curr + self.eps
                    )
                    x_new = torch.clamp(x_new, 0.0, 1.0)
                    x_new = self.check_shape(x_new)

                    margin, loss = self.margin_and_loss(x_new, y_curr)

                    # update loss if new loss is better
                    idx_improved = (loss < loss_min_curr).float()

                    loss_min[idx_to_fool] = (
                        idx_improved * loss + (1.0 - idx_improved) * loss_min_curr
                    )

                    # update margin and x_best if new loss is better
                    # or misclassification
                    idx_miscl = (margin <= 0.0).float()
                    idx_improved = torch.max(idx_improved, idx_miscl)

                    margin_min[idx_to_fool] = (
                        idx_improved * margin + (1.0 - idx_improved) * margin_min_curr
                    )
                    idx_improved = idx_improved.reshape([-1, *[1] * len(x.shape[:-1])])
                    x_best[idx_to_fool] = (
                        idx_improved * x_new + (1.0 - idx_improved) * x_best_curr
                    )
                    n_queries[idx_to_fool] += 1.0

                    ind_succ = (margin_min <= 0.0).nonzero().squeeze()
                    if self.verbose and ind_succ.numel() != 0:
                        print(
                            "{}".format(i_iter + 1),
                            "- success rate={}/{} ({:.2%})".format(
                                ind_succ.numel(),
                                n_ex_total,
                                float(ind_succ.numel()) / n_ex_total,
                            ),
                            "- avg # queries={:.1f}".format(
                                n_queries[ind_succ].mean().item()
                            ),
                            "- med # queries={:.1f}".format(
                                n_queries[ind_succ].median().item()
                            ),
                            "- loss={:.3f}".format(loss_min.mean()),
                        )

                    if ind_succ.numel() == n_ex_total:
                        break

            elif self.norm == "L2":
                delta_init = torch.zeros_like(x)
                s = h // 5
                sp_init = (h - s * 5) // 2
                vh = sp_init + 0
                for _ in range(h // s):
                    vw = sp_init + 0
                    for _ in range(w // s):
                        delta_init[:, :, vh : vh + s, vw : vw + s] += self.eta(s).view(
                            1, 1, s, s
                        ) * self.random_choice([x.shape[0], c, 1, 1])
                        vw += s
                    vh += s

                x_best = torch.clamp(
                    x + self.normalize_delta(delta_init) * self.eps, 0.0, 1.0
                )
                margin_min, loss_min = self.margin_and_loss(x_best, y)
                n_queries = torch.ones(x.shape[0]).to(self.device)
                s_init = int(math.sqrt(self.p_init * n_features / c))

                for i_iter in range(self.n_queries):
                    idx_to_fool = (margin_min > 0.0).nonzero().flatten()

                    if len(idx_to_fool) == 0:
                        break

                    x_curr = self.check_shape(x[idx_to_fool])
                    x_best_curr = self.check_shape(x_best[idx_to_fool])
                    y_curr = y[idx_to_fool]
                    if len(y_curr.shape) == 0:
                        y_curr = y_curr.unsqueeze(0)
                    margin_min_curr = margin_min[idx_to_fool]
                    loss_min_curr = loss_min[idx_to_fool]

                    delta_curr = x_best_curr - x_curr
                    p = self.p_selection(i_iter)
                    s = max(int(round(math.sqrt(p * n_features / c))), 3)
                    if s % 2 == 0:
                        s += 1

                    vh = self.random_int(0, h - s)
                    vw = self.random_int(0, w - s)
                    new_deltas_mask = torch.zeros_like(x_curr)
                    new_deltas_mask[:, :, vh : vh + s, vw : vw + s] = 1.0
                    norms_window_1 = (
                        (delta_curr[:, :, vh : vh + s, vw : vw + s] ** 2)
                        .sum(dim=(-2, -1), keepdim=True)
                        .sqrt()
                    )

                    vh2 = self.random_int(0, h - s)
                    vw2 = self.random_int(0, w - s)
                    new_deltas_mask_2 = torch.zeros_like(x_curr)
                    new_deltas_mask_2[:, :, vh2 : vh2 + s, vw2 : vw2 + s] = 1.0

                    norms_image = self.lp_norm(x_best_curr - x_curr)
                    mask_image = torch.max(new_deltas_mask, new_deltas_mask_2)
                    norms_windows = self.lp_norm(delta_curr * mask_image)

                    new_deltas = torch.ones([x_curr.shape[0], c, s, s]).to(self.device)
                    new_deltas *= self.eta(s).view(1, 1, s, s) * self.random_choice(
                        [x_curr.shape[0], c, 1, 1]
                    )
                    old_deltas = delta_curr[:, :, vh : vh + s, vw : vw + s] / (
                        1e-12 + norms_window_1
                    )
                    new_deltas += old_deltas
                    new_deltas = (
                        new_deltas
                        / (
                            1e-12
                            + (new_deltas ** 2).sum(dim=(-2, -1), keepdim=True).sqrt()
                        )
                        * (
                            torch.max(
                                (self.eps * torch.ones_like(new_deltas)) ** 2
                                - norms_image ** 2,
                                torch.zeros_like(new_deltas),
                            )
                            / c
                            + norms_windows ** 2
                        ).sqrt()
                    )
                    delta_curr[:, :, vh2 : vh2 + s, vw2 : vw2 + s] = 0.0
                    delta_curr[:, :, vh : vh + s, vw : vw + s] = new_deltas + 0

                    x_new = torch.clamp(
                        x_curr + self.normalize_delta(delta_curr) * self.eps, 0.0, 1.0
                    )
                    x_new = self.check_shape(x_new)
                    norms_image = self.lp_norm(x_new - x_curr)

                    margin, loss = self.margin_and_loss(x_new, y_curr)

                    # update loss if new loss is better
                    idx_improved = (loss < loss_min_curr).float()

                    loss_min[idx_to_fool] = (
                        idx_improved * loss + (1.0 - idx_improved) * loss_min_curr
                    )

                    # update margin and x_best if new loss is better
                    # or misclassification
                    idx_miscl = (margin <= 0.0).float()
                    idx_improved = torch.max(idx_improved, idx_miscl)

                    margin_min[idx_to_fool] = (
                        idx_improved * margin + (1.0 - idx_improved) * margin_min_curr
                    )
                    idx_improved = idx_improved.reshape([-1, *[1] * len(x.shape[:-1])])
                    x_best[idx_to_fool] = (
                        idx_improved * x_new + (1.0 - idx_improved) * x_best_curr
                    )
                    n_queries[idx_to_fool] += 1.0

                    ind_succ = (margin_min <= 0.0).nonzero().squeeze()
                    if self.verbose and ind_succ.numel() != 0:
                        print(
                            "{}".format(i_iter + 1),
                            "- success rate={}/{} ({:.2%})".format(
                                ind_succ.numel(),
                                n_ex_total,
                                float(ind_succ.numel()) / n_ex_total,
                            ),
                            "- avg # queries={:.1f}".format(
                                n_queries[ind_succ].mean().item()
                            ),
                            "- med # queries={:.1f}".format(
                                n_queries[ind_succ].median().item()
                            ),
                            "- loss={:.3f}".format(loss_min.mean()),
                        )

                    assert (x_new != x_new).sum() == 0
                    assert (x_best != x_best).sum() == 0

                    if ind_succ.numel() == n_ex_total:
                        break

        return n_queries, x_best

    def perturb(self, x, y=None):
        """
        :param x:           clean images
        :param y:           untargeted attack -> clean labels,
                            if None we use the predicted labels
                            targeted attack -> target labels, if None random classes,
                            different from the predicted ones, are sampled
        """

        self.init_hyperparam(x)

        adv = x.clone()
        if y is None:
            if not self.targeted:
                with torch.no_grad():
                    output = self.get_logits(x)
                    y_pred = output.max(1)[1]
                    y = y_pred.detach().clone().long().to(self.device)
            else:
                with torch.no_grad():
                    y = self.get_target_label(x, None)
        else:
            if not self.targeted:
                y = y.detach().clone().long().to(self.device)
            else:
                y = self.get_target_label(x, y)

        if not self.targeted:
            acc = self.get_logits(x).max(1)[1] == y
        else:
            acc = self.get_logits(x).max(1)[1] != y

        startt = time.time()

        torch.random.manual_seed(self.seed)
        torch.cuda.random.manual_seed(self.seed)

        for counter in range(self.n_restarts):
            ind_to_fool = acc.nonzero().squeeze()
            if len(ind_to_fool.shape) == 0:
                ind_to_fool = ind_to_fool.unsqueeze(0)
            if ind_to_fool.numel() != 0:
                x_to_fool = x[ind_to_fool].clone()
                y_to_fool = y[ind_to_fool].clone()

                _, adv_curr = self.attack_single_run(x_to_fool, y_to_fool)

                output_curr = self.get_logits(adv_curr)
                if not self.targeted:
                    acc_curr = output_curr.max(1)[1] == y_to_fool
                else:
                    acc_curr = output_curr.max(1)[1] != y_to_fool
                ind_curr = (acc_curr == 0).nonzero().squeeze()

                acc[ind_to_fool[ind_curr]] = 0
                adv[ind_to_fool[ind_curr]] = adv_curr[ind_curr].clone()
                if self.verbose:
                    print(
                        "restart {} - robust accuracy: {:.2%}".format(
                            counter, acc.float().mean()
                        ),
                        "- cum. time: {:.1f} s".format(time.time() - startt),
                    )

        return adv



import torch
import torch.nn.functional as F
import copy

from ..attack import Attack


class Noise:
    def __init__(self, noise_type, noise_sd):
        self.noise_type = noise_type
        self.noise_sd = noise_sd

    def __call__(self, img):
        if self.noise_type == "guassian":
            noise = torch.randn_like(img.float()) * self.noise_sd
        elif self.noise_type == "uniform":
            noise = (torch.rand_like(img.float()) - 0.5) * 2 * self.noise_sd
        return noise


class PGDRS(Attack):
    r"""
    PGD for randmized smoothing in the paper 'Provably Robust Deep Learning via Adversarially Trained Smoothed Classifiers'
    [https://arxiv.org/abs/1906.04584]
    Modification of the code from https://github.com/Hadisalman/smoothing-adversarial/blob/master/code/attacks.py

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        steps (int): number of steps. (Default: 40)
        noise_type (str): guassian or uniform. (Default: guassian)
        noise_sd (float): standard deviation for normal distributio, or range for . (Default: 0.5)
        noise_batch_size (int): guassian or uniform. (Default: 5)
        batch_max (int): split data into small chunk if the total number of augmented data points, len(inputs)*noise_batch_size, are larger than batch_max, in case GPU memory is insufficient. (Default: 2048)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.PGDRS(model, eps=8/255, alpha=2/255, steps=10, noise_type="guassian", noise_sd=0.5, noise_batch_size=5, batch_max=2048)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        eps=8 / 255,
        alpha=2 / 255,
        steps=10,
        noise_type="guassian",
        noise_sd=0.5,
        noise_batch_size=5,
        batch_max=2048,
    ):
        super().__init__("PGDRS", model)
        self.eps = eps
        self.alpha = alpha
        self.steps = steps
        self.noise_func = Noise(noise_type, noise_sd)
        self.noise_batch_size = noise_batch_size
        self.supported_mode = ["default", "targeted"]
        self.batch_max = batch_max

    def forward(self, inputs: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        if inputs.shape[0] * self.noise_batch_size > self.batch_max:
            split_num = int(self.batch_max / self.noise_batch_size)
            inputs_split = torch.split(inputs, split_size_or_sections=split_num)
            labels_split = torch.split(labels, split_size_or_sections=split_num)
            img_list = []
            for img_sub, lab_sub in zip(inputs_split, labels_split):
                img_adv = self._forward(img_sub, lab_sub)
                img_list.append(img_adv)
            return torch.vstack(img_list)
        else:
            return self._forward(inputs, labels)

    def _forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)
        # expend the inputs over noise_batch_size
        shape = (
            torch.Size([images.shape[0], self.noise_batch_size]) + images.shape[1:]
        )  # nopep8
        inputs_exp = images.unsqueeze(1).expand(shape)
        inputs_exp = inputs_exp.reshape(
            torch.Size([-1]) + inputs_exp.shape[2:]
        )  # nopep8

        delta = torch.zeros(
            (len(labels), *inputs_exp.shape[1:]), requires_grad=True, device=self.device
        )
        delta_last = torch.zeros(
            (len(labels), *inputs_exp.shape[1:]),
            requires_grad=False,
            device=self.device,
        )

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        for _ in range(self.steps):
            delta.requires_grad = True
            # img_adv is the perturbed data for randmized smoothing
            # delta.repeat(1,self.noise_batch_size,1,1).view_as(inputs_exp)
            img_adv = inputs_exp + delta.unsqueeze(1).repeat(
                (1, self.noise_batch_size, 1, 1, 1)
            ).view_as(
                inputs_exp
            )  # nopep8
            img_adv = torch.clamp(img_adv, min=0, max=1)

            noise_added = self.noise_func(img_adv.view(len(img_adv), -1))
            noise_added = noise_added.view(img_adv.shape)

            adv_noise = img_adv + noise_added
            logits = self.get_logits(adv_noise)

            softmax = F.softmax(logits, dim=1)
            # average the probabilities across noise
            average_softmax = (
                softmax.reshape(-1, self.noise_batch_size, logits.shape[-1])
                .mean(1, keepdim=True)
                .squeeze(1)
            )
            logsoftmax = torch.log(average_softmax.clamp(min=1e-20))
            ce_loss = (
                F.nll_loss(logsoftmax, labels)
                if not self.targeted
                else -F.nll_loss(logsoftmax, target_labels)
            )

            grad = torch.autograd.grad(
                ce_loss, delta, retain_graph=False, create_graph=False
            )[0]
            delta = delta_last + self.alpha * torch.sign(grad)
            delta = torch.clamp(delta, min=-self.eps, max=self.eps)
            delta_last.data = copy.deepcopy(delta.data)

        adv_images = torch.clamp(images + delta, min=0, max=1).detach()
        return adv_images



import torch
import torch.nn as nn

from ..attack import Attack


class FFGSM(Attack):
    r"""
    New FGSM proposed in 'Fast is better than free: Revisiting adversarial training'
    [https://arxiv.org/abs/2001.03994]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 10/255)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.FFGSM(model, eps=8/255, alpha=10/255)
        >>> adv_images = attack(images, labels)
    """

    def __init__(self, model, eps=8 / 255, alpha=10 / 255):
        super().__init__("FFGSM", model)
        self.eps = eps
        self.alpha = alpha
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        loss = nn.CrossEntropyLoss()

        adv_images = images + torch.randn_like(images).uniform_(
            -self.eps, self.eps
        )  # nopep8
        adv_images = torch.clamp(adv_images, min=0, max=1).detach()
        adv_images.requires_grad = True

        outputs = self.get_logits(adv_images)

        # Calculate loss
        if self.targeted:
            cost = -loss(outputs, target_labels)
        else:
            cost = loss(outputs, labels)

        # Update adversarial images
        grad = torch.autograd.grad(
            cost, adv_images, retain_graph=False, create_graph=False
        )[0]

        adv_images = adv_images + self.alpha * grad.sign()
        delta = torch.clamp(adv_images - images, min=-self.eps, max=self.eps)
        adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

import ssl

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
from torch.autograd import Variable
from torchvision import transforms

from airtk.data import CIFAR100
from airtk.model import (
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
    WideResNet,
)


def eval_clean(model, device, data_loader, name, ds_name, f):
    """

    Calculates the natural error of the model on either the training
    or test set of data


    :param model: trained model to be evaluated

    :param device: the device the model is set on

    :param data_loader: data loader containing the dataset to test

    :param name: 'train' or 'test', denoting which dataset is being tested

    :param ds_name: name of dataset, cifar10 or cifar100

    :param f: file to print results to

    """

    model.eval()

    # Set the total errors to 0 for all attacks but AA

    total_err = 0

    # Run tests for each element in the test set

    for sample in data_loader:
        # Set up the data/X and target/y correctly for evaluation

        if ds_name == "cifar100":
            data, target, _ = sample

        else:
            data, target = sample

        data, target = data.to(device), target.to(device)

        X, y = Variable(data, requires_grad=True), Variable(target)

        # Calculate the natural error for each attack

        err = clean(model, X, y)

        # Add the losses to the total loss for each attack

        total_err += err

    # Write the total losses to the file

    if name == "test":
        # Convert the clean loss to clean accuracy %
        clean_total = int(total_err)

        clean_acc = (10000 - clean_total) / 100

        print("Clean Accuracy (Test): " + str(clean_acc) + "%")

        f.write("Clean Accuracy (Test): " + str(clean_acc) + "%\n")

    elif name == "train":
        # Convert the clean loss to clean accuracy %
        clean_total = int(total_err)

        clean_acc = (50000 - clean_total) / 500

        print("Clean Accuracy (Train): " + str(clean_acc) + "%")

        f.write("Clean Accuracy (Train): " + str(clean_acc) + "%\n")

    else:
        raise NotImplementedError


def robust_eval(model, device, test_loader, ds_name, f):
    """

    Calculates the clean loss and robust  loss against PGD and FGSM attacks


    :param model: trained model to be evaluated

    :param device: the device the model is set on

    :param test_loader: data loader containing the testing dataset

    :param f: the file to write results to


    :return pgd_acc: the accuracy against PGD attack

    """

    model.eval()

    # Set the total errors to 0 for all atts but AA

    clean_total = 0

    pgd_robust_total = 0

    fgsm_robust_total = 0

    # Run tests for each element in the test set

    for sample in test_loader:
        # Set up the data/X and target/y correctly for evaluation

        if ds_name == "cifar100":
            data, target, _ = sample

        else:
            data, target = sample

        data, target = data.to(device), target.to(device)

        X, y = Variable(data, requires_grad=True), Variable(target)

        # Calculate the natural and robust error for each attack

        pgd_err_natural, pgd_err_robust = pgd_whitebox_eval(
            model, X, y, device
        )

        fgsm_err_natural, fgsm_err_robust = fgsm_whitebox_eval(model, X, y)

        # Add the losses to the total loss for each attack

        clean_total += pgd_err_natural

        pgd_robust_total += pgd_err_robust

        fgsm_robust_total += fgsm_err_robust

    # Convert the clean error to clean accuracy %
    clean_total = int(clean_total)

    clean_acc = (10000 - clean_total) / 100

    # Convert the PGD error to clean accuracy %

    pgd_robust_total = int(pgd_robust_total)

    pgd_acc = (10000 - pgd_robust_total) / 100

    # Convert the FGSM error to clean accuracy %

    fgsm_robust_total = int(fgsm_robust_total)

    fgsm_acc = (10000 - fgsm_robust_total) / 100

    # Print out the loss percents

    print("Clean Accuracy (Test): " + str(clean_acc) + "%")

    print("PGD Robust Accuracy: " + str(pgd_acc) + "%")

    print("FGSM Robust Accuracy: " + str(fgsm_acc) + "%")

    # Write loss percents to file

    f.write("Clean Accuracy (Test): " + str(clean_acc) + "%\n")

    f.write("PGD Robust Accuracy: " + str(pgd_acc) + "%\n")

    f.write("FGSM Robust Accuracy: " + str(fgsm_acc) + "%\n")

    return pgd_acc


def adjust_learning_rate(optimizer, epoch, args):
    """

    Sets the learning rate of the optimizer based on the current epoch


    :param optimizer: optimizer with learning rate being set

    :param epoch: current epoch

    :param args: program arguments

    """
    lr = args.lr

    if args.lr_schedule == "decay":
        if epoch >= 50:
            lr = args.lr * 0.1

        if epoch >= 75:
            lr = args.lr * 0.01

    elif args.lr_schedule == "scheduled":
        if epoch >= 24:
            lr = args.lr * 0.1

        if epoch >= 26:
            lr = args.lr

        if epoch >= 64:
            lr = args.lr * 0.1

        if epoch >= 66:
            lr = args.lr

        if epoch >= 104:
            lr = args.lr * 0.1

        if epoch >= 106:
            lr = args.lr

        if epoch >= 137:
            lr = args.lr * 0.1

        if epoch >= 139:
            lr = args.lr * 0.01

    elif args.lr_schedule == "cosine":
        lr = 0.2

        lr = lr * 0.5 * (1 + np.cos((epoch - 1) / args.epochs * np.pi))

    elif args.lr_schedule == "cyclic-10":
        if epoch % 10 < 2 or epoch % 10 == 9:
            lr = args.lr

        elif epoch % 10 < 4 or epoch % 10 > 6:
            lr = args.lr * 0.1

        elif epoch % 10 == 5:
            lr = args.lr * 0.001

        else:
            lr = args.lr * 0.01

    elif args.lr_schedule == "cyclic-5":
        if epoch % 5 == 0:
            lr = args.lr

        elif epoch % 5 == 1 or epoch % 5 == 4:
            lr = args.lr * 0.1

        else:
            lr = args.lr * 0.01

    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def load_data(ds_name, args, kwargs, coarse=False):
    """

    Loads the specified dataset into a dataloader and returns the data split into
    training and test loaders


    :param ds_name: name of the dataset to load for training

    :param args: program arguments

    :param kwargs: more program arguments


    :return data_loaders containing the training and testing sets

    """

    # Establish the data loader transforms

    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ]
    )

    transform_test = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )

    ssl._create_default_https_context = (
        ssl._create_unverified_context
    )  # set the context for working with tensors

    if ds_name == "cifar10":
        # Load in the CIFAR10 dataloaders

        trainset = torchvision.datasets.CIFAR10(
            root="../data/download",
            train=True,
            download=True,
            transform=transform_train,
        )

        train_loader = torch.utils.data.DataLoader(
            trainset, batch_size=args.batch_size, shuffle=True, **kwargs
        )

        testset = torchvision.datasets.CIFAR10(
            root="../data/download",
            train=False,
            download=True,
            transform=transform_test,
        )

        test_loader = torch.utils.data.DataLoader(
            testset, batch_size=args.test_batch_size, shuffle=False, **kwargs
        )

    elif ds_name == "cifar100":
        # Load in the CIFAR100 dataloaders

        if coarse:
            trainset = CIFAR100(
                root="../data/download",
                train=True,
                download=True,
                transform=transform_train,
                coarse=True,
            )

            train_loader = torch.utils.data.DataLoader(
                trainset, batch_size=args.batch_size, shuffle=True, **kwargs
            )

            testset = CIFAR100(
                root="../data/download",
                train=False,
                download=True,
                transform=transform_test,
                coarse=True,
            )

            test_loader = torch.utils.data.DataLoader(
                testset,
                batch_size=args.test_batch_size,
                shuffle=False,
                **kwargs,
            )

        else:
            trainset = torchvision.datasets.CIFAR100(
                root="../data/download",
                train=True,
                download=True,
                transform=transform_train,
            )

            train_loader = torch.utils.data.DataLoader(
                trainset, batch_size=args.batch_size, shuffle=True, **kwargs
            )

            testset = torchvision.datasets.CIFAR100(
                root="../data/download",
                train=False,
                download=True,
                transform=transform_test,
            )

            test_loader = torch.utils.data.DataLoader(
                testset,
                batch_size=args.test_batch_size,
                shuffle=False,
                **kwargs,
            )

    else:
        raise NotImplementedError

    return train_loader, test_loader


def class_define_attacks(dataset):
    """

    Defines a dictionary of attacks based on the image-class.


    :param epoch: current epoch

    :param dataset: name of dataset


    :return attacks: dictionary of attacks

    """

    attacks = {}

    if dataset == "cifar10":
        attacks[0] = "mim"

        attacks[1] = "linf-pgd-20"

        attacks[2] = "linf-pgd-40"

        attacks[3] = "cw"

        attacks[4] = "apgd-ce"

        attacks[5] = "linf-pgd-20"

        attacks[6] = "linf-pgd-7"

        attacks[7] = "apgd-dlr"

        attacks[8] = "apgd-t"

        attacks[9] = "linf-pgd-40"

    elif dataset == "cifar100":
        attacks[0] = "mim"

        attacks[1] = "linf-pgd-20"

        attacks[2] = "linf-pgd-40"

        attacks[3] = "cw"

        attacks[4] = "apgd-ce"

        attacks[5] = "linf-pgd-20"

        attacks[6] = "linf-pgd-7"

        attacks[7] = "apgd-dlr"

        attacks[8] = "apgd-t"

        attacks[9] = "linf-pgd-40"

        attacks[10] = "mim"

        attacks[11] = "linf-pgd-20"

        attacks[12] = "linf-pgd-40"

        attacks[13] = "cw"

        attacks[14] = "apgd-ce"

        attacks[15] = "linf-pgd-20"

        attacks[16] = "linf-pgd-7"

        attacks[17] = "apgd-dlr"

        attacks[18] = "apgd-t"

        attacks[19] = "linf-pgd-40"

    else:
        raise NotImplementedError

    return attacks


def one_hot_tensor(y_batch_tensor, num_classes, device):
    """

    Converts a batch tensor into a one-hot tensor


    :param y_batch_tensor: batch tensor to be converted

    :param num_classes: number of classes to fill the tensor

    :param device: current device


    :return one-hot tensor based on the input batch tensor

    """

    y_tensor = torch.cuda.FloatTensor(
        y_batch_tensor.size(0), num_classes
    ).fill_(0)

    y_tensor[np.arange(len(y_batch_tensor)), y_batch_tensor] = 1.0

    return y_tensor


class CWLoss(nn.Module):
    """Class for the CW loss, including parameters as well as a method to propogate

    the loss forwards"""

    def __init__(self, num_classes, margin=50, reduce=True):
        """

        Set up the CW loss with number of classes, margin of 50, and reduce = True

        """

        super(CWLoss, self).__init__()
        self.num_classes = num_classes
        self.margin = margin
        self.reduce = reduce
        return

    def forward(self, logits, targets):
        """

        :param logits: predictions

        :param targets: target labels

        :return: loss

        """

        # Convert the target labels to a one-hot tensor
        onehot_targets = one_hot_tensor(
            targets, self.num_classes, targets.device
        )

        # Calculate self-loss (sum of targets/predictions)

        self_loss = torch.sum(onehot_targets * logits, dim=1)

        # Calculate other-loss

        other_loss = torch.max(
            (1 - onehot_targets) * logits - onehot_targets * 1000, dim=1
        )[0]

        # Take the loss as the reverse sum of the loss differences plus the margin (clamped)

        loss = -torch.sum(torch.clamp(self_loss - other_loss + self.margin, 0))

        # If reducing (default True), divide the loss by the number of targets

        if self.reduce:
            sample_num = onehot_targets.shape[0]

            loss = loss / sample_num

        return loss


def clean(model, X, y):
    """

    Evaluates a model on a clean sample.


    :param model: classifier to be evaluated

    :param X: image

    :param y: correct classification of image


    :return the error between the model prediction of the image and the correct classification

    """

    out = model(X)  # model prediction
    err = (
        (out.data.max(1)[1] != y.data).float().sum()
    )  # error between prediction and classification
    return err


def pgd_whitebox_eval(
    model, X, y, device, epsilon=0.031, num_steps=20, step_size=2.0 / 255.0
):
    """

    Evaluates the model by perturbing an image using the PGD attack.


    :param model: model being attacked

    :param X: image being attacked

    :param y: correct label of the image being attacked

    :param device: current device

    :param epsilon: epsilon size for the PGD attack

    :param num_steps: number of perturbation steps

    :param step_size: step size of perturbation steps


    :return clean error and PGD error

    """

    # Calculates clean error of image classification

    out = model(X)

    err = (out.data.max(1)[1] != y.data).float().sum()

    # Create X_pgd basis by duplicating X as a variable

    X_pgd = Variable(X.data, requires_grad=True)

    # If specified, create random noice between - and + epsilon and add to X_pgd
    random_noise = (
        torch.FloatTensor(*X_pgd.shape).uniform_(-epsilon, epsilon).to(device)
    )

    X_pgd = Variable(X_pgd.data + random_noise, requires_grad=True)

    # For each perturbation step

    for _ in range(num_steps):
        # Create the SGD optimizer and zero the gradients

        opt = optim.SGD([X_pgd], lr=1e-3)

        opt.zero_grad()

        # With gradients, set up the cross entropy loss and step backward

        with torch.enable_grad():
            loss = nn.CrossEntropyLoss()(model(X_pgd), y)

        loss.backward()

        # Calculate the perturbation eta as the step size in the gradient direction of X_pgd

        eta = step_size * X_pgd.grad.data.sign()

        # Add the perturbation to X_pgd

        X_pgd = Variable(X_pgd.data + eta, requires_grad=True)

        # Set the perturbation to the difference between X and X_adv, clamped by +/- epsilon

        eta = torch.clamp(X_pgd.data - X.data, -epsilon, epsilon)

        # Add the new perturbation to X_pgd again

        X_pgd = Variable(X.data + eta, requires_grad=True)

        # Clamp X_gd to be between 0 and 1, as a Variable

        X_pgd = Variable(torch.clamp(X_pgd, 0, 1.0), requires_grad=True)

    # Calculate the error between the PGD-perturbed prediction and correct classification

    err_pgd = (model(X_pgd).data.max(1)[1] != y.data).float().sum()

    return err, err_pgd


def fgsm_whitebox_eval(model, X, y, epsilon=0.031):
    """

    Evaluates the model by perturbing an image using the FGSM attack.


    :param model: model being attacked

    :param X: image being attacked

    :param y: correct label of the image being attacked

    :param epsilon: epsilon size for the FGSM attack


    :return clean error and FGSM error

    """

    # Calculates clean error of image classification

    out = model(X)

    err = (out.data.max(1)[1] != y.data).float().sum()

    # Create X_fgsm basis by duplicating X as a variable

    X_fgsm = Variable(X.data, requires_grad=True)

    # Create the SGD optimizer and zero the gradients

    opt = optim.SGD([X_fgsm], lr=1e-3)

    opt.zero_grad()

    # With gradients, set up the cross entropy loss and step backward

    with torch.enable_grad():
        loss = nn.CrossEntropyLoss()(model(X_fgsm), y)

    loss.backward()

    # Discover X_fgsm by adding epsilon in the gradient direction and clamping the sample

    X_fgsm = Variable(
        torch.clamp(X_fgsm.data + epsilon * X_fgsm.grad.data.sign(), 0.0, 1.0),
        requires_grad=True,
    )

    # Calculates the FGSM error between the prediction and correct classification

    err_fgsm = (model(X_fgsm).data.max(1)[1] != y.data).float().sum()

    return err, err_fgsm


def cw_whitebox_eval(
    model,
    dataset,
    X,
    y,
    device,
    epsilon=0.031,
    num_steps=20,
    step_size=2.0 / 255.0,
):
    """

    Evaluates the model by perturbing an image using the CW attack.


    :param model: model being attacked

    :param dataset: name of dataset (cifar10 or cifar100)

    :param X: image being attacked

    :param y: correct label of the image being attacked

    :param device: current device

    :param epsilon: epsilon size for the CW attack

    :param num_steps: number of perturbation steps

    :param step_size: step size of perturbation steps


    :return clean error and CW error

    """

    # Calculates clean error of image classification

    out = model(X)

    err = (out.data.max(1)[1] != y.data).float().sum()

    # Create X_cw basis by duplicating X as a variable

    X_cw = Variable(X.data, requires_grad=True)

    # If specified, create random noice between - and + epsilon and add to X_cw
    random_noise = (
        torch.FloatTensor(*X_cw.shape).uniform_(-epsilon, epsilon).to(device)
    )

    X_cw = Variable(X_cw.data + random_noise, requires_grad=True)

    for _ in range(num_steps):
        # Create the SGD optimizer and zero the gradients

        opt = optim.SGD([X_cw], lr=1e-3)

        opt.zero_grad()

        # With gradients, set up the CW loss and step backward

        with torch.enable_grad():
            loss = CWLoss(100 if dataset == "cifar100" else 10)(model(X_cw), y)

        loss.backward()

        # Calculate the perturbation using the CW loss gradient

        eta = step_size * X_cw.grad.data.sign()

        X_cw = Variable(X_cw.data + eta, requires_grad=True)

        eta = torch.clamp(X_cw.data - X.data, -epsilon, epsilon)

        # Add the perturbation and clamp x_cw

        X_cw = Variable(X.data + eta, requires_grad=True)

        X_cw = Variable(torch.clamp(X_cw, 0, 1.0), requires_grad=True)

    # Calculates the CW error between the prediction and correct classification

    err_cw = (model(X_cw).data.max(1)[1] != y.data).float().sum()

    return err, err_cw


def mim_whitebox_eval(
    model,
    X,
    y,
    device,
    epsilon=0.031,
    num_steps=20,
    step_size=0.031,
    decay_factor=1.0,
):
    """

    Evaluates the model by perturbing an image using the MIM attack.


    :param model: model being attacked

    :param X: image being attacked

    :param y: correct label of the image being attacked

    :param device: current device

    :param epsilon: epsilon size for the MIM attack

    :param num_steps: number of perturbation steps

    :param step_size: step size of perturbation steps

    :param decay_factor: coefficient for previous gradient usage in MIM attack


    :return clean error and MIM error

    """

    # Calculates clean error of image classification

    out = model(X)

    err = (out.data.max(1)[1] != y.data).float().sum()

    # Create X_mim basis by duplicating X as a variable

    X_mim = Variable(X.data, requires_grad=True)

    # create random noise between - and + epsilon and add to X_cw
    random_noise = (
        torch.FloatTensor(*X_mim.shape).uniform_(-epsilon, epsilon).to(device)
    )

    X_mim = Variable(X_mim.data + random_noise, requires_grad=True)

    # Set up tensor to hold previous gradients

    previous_grad = torch.zeros_like(X.data)

    for _ in range(num_steps):
        # Create the SGD optimizer and zero the gradients

        opt = optim.SGD([X_mim], lr=1e-3)

        opt.zero_grad()

        # With gradients, set up the cross entropy loss and step backward

        with torch.enable_grad():
            loss = nn.CrossEntropyLoss()(model(X_mim), y)

        loss.backward()

        # Calculate the perturbation using the current and previous gradient

        grad = X_mim.grad.data / torch.mean(
            torch.abs(X_mim.grad.data), [1, 2, 3], keepdim=True
        )

        previous_grad = decay_factor * previous_grad + grad

        X_mim = Variable(
            X_mim.data + step_size * previous_grad.sign(), requires_grad=True
        )

        eta = torch.clamp(X_mim.data - X.data, -epsilon, epsilon)

        # Add the perturbation and clamp x_mim

        X_mim = Variable(X.data + eta, requires_grad=True)

        X_mim = Variable(torch.clamp(X_mim, 0, 1.0), requires_grad=True)

    # Calculates the MIM error between the prediction and correct classification

    err_mim = (model(X_mim).data.max(1)[1] != y.data).float().sum()

    return err, err_mim


def cw_whitebox(
    model,
    X,
    y,
    device,
    dataset,
    epsilon=0.031,
    num_steps=20,
    step_size=2.0 / 255.0,
):
    """

    Attacks the specified image X using the CW attack and returns the adversarial example


    :param model: model being attacked

    :param X: image being attacked

    :param y: correct label of the image being attacked

    :param device: current device

    :param epsilon: epsilon size for the PGD attack

    :param num_steps: number of perturbation steps

    :param step_size: step size of perturbation steps


    :return adversarial example found with the CW attack

    """

    # Create X_pgd basis by duplicating X as a variable

    X_pgd = Variable(X.data, requires_grad=True)

    # If adding random, create random noice between - and + epsilon and add to X_pgd
    random_noise = (
        torch.FloatTensor(*X_pgd.shape).uniform_(-epsilon, epsilon).to(device)
    )

    X_pgd = Variable(X_pgd.data + random_noise, requires_grad=True)

    # For each perturbation step:

    for _ in range(num_steps):
        # Create the SGD optimizer and zero the gradients

        opt = optim.SGD([X_pgd], lr=1e-3)

        opt.zero_grad()

        # With gradients, set up the CW loss and step backward

        with torch.enable_grad():
            loss = CWLoss(100 if dataset == "cifar100" else 10)(
                model(X_pgd), y
            )

        loss.backward()

        # Calculate the perturbation eta as the step size in the gradient direction of X_pgd

        eta = step_size * X_pgd.grad.data.sign()

        # Add the perturbation to X_pgd

        X_pgd = Variable(X_pgd.data + eta, requires_grad=True)

        # Set the perturbation to the difference between X and X_adv, clamped by +/- epsilon

        eta = torch.clamp(X_pgd.data - X.data, -epsilon, epsilon)

        # Add the new perturbation to X_pgd again

        X_pgd = Variable(X.data + eta, requires_grad=True)

        # Clamp X_gd to be between 0 and 1, as a Variable

        X_pgd = Variable(torch.clamp(X_pgd, 0, 1.0), requires_grad=True)

    return X_pgd


def mim_whitebox(
    model,
    X,
    y,
    device,
    epsilon=0.031,
    num_steps=20,
    step_size=2.0 / 255.0,
    decay_factor=1.0,
):
    """

    Attacks the specified image X using the MIM attack and returns the adversarial example


    :param model: model being attacked

    :param X: image being attacked

    :param y: correct label of the image being attacked

    :param device: current device

    :param epsilon: epsilon size for the PGD attack

    :param num_steps: number of perturbation steps

    :param step_size: step size of perturbation steps

    :param decay_factor: factor of decay for gradients


    :return adversarial example found with the MIM attack

    """

    # Create X_pgd basis by duplicating X as a variable

    X_pgd = Variable(X.data, requires_grad=True)

    # If adding random, create random noice between - and + epsilon and add to X_pgd
    random_noise = (
        torch.FloatTensor(*X_pgd.shape).uniform_(-epsilon, epsilon).to(device)
    )

    X_pgd = Variable(X_pgd.data + random_noise, requires_grad=True)

    # Set the previous gradient to a tensor of 0s in the shape of X

    previous_grad = torch.zeros_like(X.data)

    # For each perturbation step:

    for _ in range(num_steps):
        # Create the SGD optimizer and zero the gradients

        opt = optim.SGD([X_pgd], lr=1e-3)

        opt.zero_grad()

        # With gradients, set up the cross-entropy loss and step backward

        with torch.enable_grad():
            loss = nn.CrossEntropyLoss()(model(X_pgd), y)

        loss.backward()

        # Calculate the gradient by dividing it by the average

        grad = X_pgd.grad.data / torch.mean(
            torch.abs(X_pgd.grad.data), [1, 2, 3], keepdim=True
        )

        # Calculate the previous gradient by multiplying it by the decay factor and adding to the grad

        previous_grad = decay_factor * previous_grad + grad

        # Perturb X_pgd in the direction of the previous grad, by the step size

        X_pgd = Variable(
            X_pgd.data + step_size * previous_grad.sign(), requires_grad=True
        )

        # Set the perturbation to the difference between X and X_adv, clamped by +/- epsilon

        eta = torch.clamp(X_pgd.data - X.data, -epsilon, epsilon)

        # Add the new perturbation to X_pgd againt

        X_pgd = Variable(X.data + eta, requires_grad=True)

        # Clamp X_gd to be between 0 and 1, as a Variable

        X_pgd = Variable(torch.clamp(X_pgd, 0, 1.0), requires_grad=True)

    return X_pgd


def get_model(mod_name, ds_name, device):
    if mod_name == "res18":
        model = ResNet18(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(
            device
        )  # TODO(Ezuharad): this style of hardcoding is not good, we should use a class attribute

    elif mod_name == "res34":
        model = ResNet34(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res50":
        model = ResNet50(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res101":
        model = ResNet101(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res152":
        model = ResNet152(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "wideres34":
        model = WideResNet(
            depth=34, num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    else:
        raise NotImplementedError

    return model



import os
from collections import namedtuple
from os import PathLike

import torch

# from tensorboardX import SummaryWriter
import torch.optim as optim

from airtk.defense.yopo.config import TrainingConfig
from airtk.defense.yopo.dataset import (
    create_test_dataset,
    create_train_dataset,
)
from airtk.defense.yopo.loss import Hamiltonian
from airtk.defense.yopo.training_function import (
    FastGradientLayerOneTrainer,
    train_one_epoch,
)
from airtk.defense.yopo.training_train import (
    eval_one_epoch,
)
from airtk.defense.yopo.utils_misc import (
    load_checkpoint,
    save_checkpoint,
)

# from network import create_network
from airtk.defense.yopo.wide_resnet import WideResNet

# torch.backends.cudnn.benchmark = True


# writer = SummaryWriter(log_dir=config.log_dir)
class YopoTraining:
    def __init__(
        self,
        dataset: str,
        epochs: int = 36,
        batch_size: int = 256,
        test_interval: int = 2,
        iters: int = 20,
        inner_iters: int = 3,
        weight_decay: float = 5e-4,
        K: int = 5,
        sigma: float = 2.0 / 255.0,
        momentum: float = 0.9,
        gamma: float = 0.1,
        eps: float = 8.0 / 255.0,
        model_dir: PathLike = "data/model",
    ) -> None:
        self._config = TrainingConfig(
            eps,
            epochs,
            test_interval,
            weight_decay,
            inner_iters,
            K,
            sigma,
            iters,
            gamma,
            momentum,
        )

        self._ds_name = dataset

        ArgPrototype = namedtuple(
            "ArgPrototype",
            ["batch_size", "resume", "auto_continue", "adv_coef", "model_dir"],
        )
        self._args = ArgPrototype(batch_size, False, False, 1.0, model_dir)

    def __call__(self) -> None:
        main_yopo(self._ds_name, self._config, self._args)


def main_yopo(ds_name, config, args):
    if ds_name == "cifar10":
        num_classes = 10
    elif ds_name == "cifar100":
        num_classes = 100
    else:
        raise NotImplementedError()
    DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
    net = WideResNet(depth=34, num_classes=num_classes).to(DEVICE)
    net.to(DEVICE)
    criterion = config.create_loss_function().to(DEVICE)
    # criterion = CrossEntropyWithWeightPenlty(net.other_layers, DEVICE, config.weight_decay)#.to(DEVICE)
    # ce_criterion = nn.CrossEntropyLoss().to(DEVICE)
    optimizer = config.create_optimizer()(net.parameters())
    lr_scheduler = config.create_lr_scheduler()(optimizer)

    ## Make Layer One trainer  This part of code should be written in config.py

    Hamiltonian_func = Hamiltonian(net.layer_one, config.weight_decay)
    layer_one_optimizer = optim.SGD(
        net.layer_one.parameters(),
        lr=lr_scheduler.get_lr()[0],
        momentum=0.9,
        weight_decay=5e-4,
    )
    lyaer_one_optimizer_lr_scheduler = optim.lr_scheduler.MultiStepLR(
        layer_one_optimizer, milestones=[30, 34, 36], gamma=0.1
    )
    LayerOneTrainer = FastGradientLayerOneTrainer(
        Hamiltonian_func,
        layer_one_optimizer,
        config.inner_iters,
        config.sigma,
        config.eps,
    )

    ds_train = create_train_dataset(ds_name, args.batch_size)
    ds_val = create_test_dataset(ds_name, args.batch_size)

    # TrainAttack = config.create_attack_method(DEVICE)
    EvalAttack = config.create_evaluation_attack_method()(DEVICE)

    now_epoch = 0

    if args.auto_continue:
        args.resume = os.path.join(args.model_dir, "last.checkpoint")
    if args.resume is not None and os.path.isfile(args.resume):
        now_epoch = load_checkpoint(args.resume, net, optimizer, lr_scheduler)

    while True:
        if now_epoch > config.num_epochs:
            break
        now_epoch = now_epoch + 1

        descrip_str = f"Training epoch:{now_epoch}/{config.num_epochs} -- lr:{lr_scheduler.get_lr()[0]}"
        acc, yofoacc = train_one_epoch(
            net,
            ds_train,
            optimizer,
            criterion,
            LayerOneTrainer,
            config.K,
            config.eps,
            DEVICE,
            descrip_str,
        )
        tb_train_dic = {"Acc": acc, "YofoAcc": yofoacc}
        print(tb_train_dic)
        # writer.add_scalars('Train', tb_train_dic, now_epoch)
        if config.val_interval > 0 and now_epoch % config.val_interval == 0:
            acc, advacc = eval_one_epoch(net, ds_val, DEVICE, EvalAttack)
            # tb_val_dic = {"Acc": acc, "AdvAcc": advacc}
            # writer.add_scalars('Val', tb_val_dic, now_epoch)

        lr_scheduler.step()
        lyaer_one_optimizer_lr_scheduler.step()
        save_checkpoint(
            now_epoch,
            net,
            optimizer,
            lr_scheduler,
            file_name=os.path.join(
                args.model_dir, f"epoch-{now_epoch}.checkpoint"
            ),
        )


# if __name__ == "__main__":
# main_yopo()



import time

from ..attack import Attack
from ..wrappers.multiattack import MultiAttack
from .apgd import APGD
from .apgdt import APGDT
from .fab import FAB
from .square import Square


class AutoAttack(Attack):
    r"""
    AutoAttack in the paper 'Reliable evaluation of adversarial robustness with an ensemble of diverse parameter-free attacks'
    [https://arxiv.org/abs/2003.01690]
    [https://github.com/fra31/auto-attack]

    Distance Measure : Linf, L2

    Arguments:
        model (nn.Module): model to attack.
        norm (str) : Lp-norm to minimize. ['Linf', 'L2'] (Default: 'Linf')
        eps (float): maximum perturbation. (Default: 0.3)
        version (bool): version. ['standard', 'plus', 'rand'] (Default: 'standard')
        n_classes (int): number of classes. (Default: 10)
        seed (int): random seed for the starting point. (Default: 0)
        verbose (bool): print progress. (Default: False)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.AutoAttack(model, norm='Linf', eps=8/255, version='standard', n_classes=10, seed=None, verbose=False)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        norm="Linf",
        eps=8 / 255,
        version="standard",
        n_classes=10,
        seed=None,
        verbose=False,
    ):
        super().__init__("AutoAttack", model)
        self.norm = norm
        self.eps = eps
        self.version = version
        self.n_classes = n_classes
        self.seed = seed
        self.verbose = verbose
        self.supported_mode = ["default"]

        if version == "standard":  # ['apgd-ce', 'apgd-t', 'fab-t', 'square']
            self._autoattack = MultiAttack(
                [
                    APGD(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        loss="ce",
                        n_restarts=1,
                    ),
                    APGDT(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        n_classes=n_classes,
                        n_restarts=1,
                    ),
                    FAB(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        multi_targeted=True,
                        n_classes=n_classes,
                        n_restarts=1,
                    ),
                    Square(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        n_queries=5000,
                        n_restarts=1,
                    ),
                ]
            )

        # ['apgd-ce', 'apgd-dlr', 'fab', 'square', 'apgd-t', 'fab-t']
        elif version == "plus":
            self._autoattack = MultiAttack(
                [
                    APGD(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        loss="ce",
                        n_restarts=5,
                    ),
                    APGD(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        loss="dlr",
                        n_restarts=5,
                    ),
                    FAB(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        n_classes=n_classes,
                        n_restarts=5,
                    ),
                    Square(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        n_queries=5000,
                        n_restarts=1,
                    ),
                    APGDT(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        n_classes=n_classes,
                        n_restarts=1,
                    ),
                    FAB(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        multi_targeted=True,
                        n_classes=n_classes,
                        n_restarts=1,
                    ),
                ]
            )

        elif version == "rand":  # ['apgd-ce', 'apgd-dlr']
            self._autoattack = MultiAttack(
                [
                    APGD(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        loss="ce",
                        eot_iter=20,
                        n_restarts=1,
                    ),
                    APGD(
                        model,
                        eps=eps,
                        norm=norm,
                        seed=self.get_seed(),
                        verbose=verbose,
                        loss="dlr",
                        eot_iter=20,
                        n_restarts=1,
                    ),
                ]
            )

        else:
            raise ValueError("Not valid version. ['standard', 'plus', 'rand']")

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)
        adv_images = self._autoattack(images, labels)

        return adv_images

    def get_seed(self):
        return time.time() if self.seed is None else self.seed



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

import argparse
import os
import time

import torch
from airtk.helper_functions import (
    adjust_learning_rate,
    eval_clean,
    load_data,
    robust_eval,
)
from airtk.loss import standard_loss
from airtk.util import get_model
from torch import optim

parser = argparse.ArgumentParser(description="PyTorch VA Adversarial Training")

parser.add_argument("--dataset", type=str, default="cifar100", help="dataset")
parser.add_argument(
    "--model-arch", default="wideres34", help="model architecture to train"
)
parser.add_argument(
    "--batch-size",
    type=int,
    default=128,
    metavar="N",
    help="input batch size for training (default: 128)",
)
parser.add_argument(
    "--test-batch-size",
    type=int,
    default=128,
    metavar="N",
    help="input batch size for testing (default: 128)",
)
parser.add_argument(
    "--epochs",
    type=int,
    default=140,
    metavar="N",
    help="number of epochs to train",
)
parser.add_argument(
    "--warmup",
    type=int,
    default=0,
    metavar="N",
    help="number of epochs to train with clean data before AT",
)
parser.add_argument(
    "--weight-decay", "--wd", default=2e-4, type=float, metavar="W"
)
parser.add_argument(
    "--lr", type=float, default=0.1, metavar="LR", help="learning rate"
)
parser.add_argument(
    "--momentum", type=float, default=0.9, metavar="M", help="SGD momentum"
)
parser.add_argument(
    "--no-cuda",
    action="store_true",
    default=False,
    help="disables CUDA training",
)
parser.add_argument(
    "--seed", type=int, default=1, metavar="S", help="random seed (default: 1)"
)
parser.add_argument(
    "--log-interval",
    type=int,
    default=100,
    metavar="N",
    help="how many batches to wait before logging training status",
)
parser.add_argument(
    "--model-dir",
    default="./data/model",
    help="directory of model for saving checkpoint",
)
parser.add_argument(
    "--save-freq",
    "-s",
    default=1,
    type=int,
    metavar="N",
    help="save frequency",
)
parser.add_argument(
    "--lr-schedule",
    default="decay",
    help="schedule for adjusting learning rate",
)
parser.add_argument(
    "--epsilon", type=float, default=8.0 / 255.0, help="perturbation"
)
parser.add_argument(
    "--num-steps", type=int, default=20, help="perturb number of steps"
)
parser.add_argument(
    "--step-size", type=float, default=2.0 / 255.0, help="perturb step size"
)
parser.add_argument(
    "--random", default=True, help="random initialization for PGD"
)

args = parser.parse_args()
model_dir = args.model_dir

if not os.path.exists(model_dir):
    os.makedirs(model_dir)

use_cuda = not args.no_cuda and torch.cuda.is_available()
torch.manual_seed(args.seed)

device = torch.device("cuda" if use_cuda else "cpu")

kwargs = {"num_workers": 1, "pin_memory": True} if use_cuda else {}


def train(args, model, device, train_loader, optimizer, epoch, ds_name):
    """

    Trains one epoch by calculating the loss for each batch and updating the model


    :param args: set of arguments including learning rate, log interval, etc

    :param model: model to train

    :param device: current device

    :param train_loader: data loader containing the training dataset

    :param optimizer: optimizer to train

    :param epoch: current epoch of training
    """

    model.train()  # set the  model to training mode

    for batch_idx, sample in enumerate(train_loader):
        # Extract the image and classification details

        if ds_name == "cifar100":
            data, target, coarse = sample

        else:
            data, target = sample

        data, target = (
            data.to(device),
            target.to(device),
        )  # set data/target to device

        optimizer.zero_grad()  # zero out the gradients

        # Calculate the clean loss

        loss, batch_metrics = standard_loss(model, data, target, optimizer)

        # Update the model based on the loss

        loss.backward()

        optimizer.step()

        # Print training progress

        if batch_idx % args.log_interval == 0:
            print(
                f"Train Epoch: {epoch} [{batch_idx * len(data)}/{len(train_loader.dataset)} ({100.0 * batch_idx / len(train_loader):.0f}%)]\tLoss: {loss.item():.6f}"
            )


def main_warmup(ds_name, mod_name, epochs):
    """

    Main training method, which establishes the model/dataset before conducting training and

    clean testing for each epoch. Then reports final training time and conducts robustness tests
    on the final model.


    :param ds_name: dataset to use for training

    :param mod_name: model to use for training

    :param epochs: number of epochs to train for
    """

    # Create file to print training progress

    filename = f"log/clean-{ds_name}-{mod_name}-output.txt"
    f = open(filename, "a")

    # Initialize the model based on the specified architecture

    start_tot = time.time()  # start recording training time

    model = get_model(mod_name, ds_name, device)

    # Set up the optimizer with the arguments

    optimizer = optim.SGD(
        model.parameters(),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    # Set up the dataloaders

    train_loader, test_loader = load_data(ds_name, args, kwargs, coarse=True)

    # Begin model training for the specified number of epochs

    for epoch in range(1, epochs + 1):
        # Adjust the SGD learning rate based on the epoch

        adjust_learning_rate(optimizer, epoch, args)

        start = time.time()  # start recording training time

        train(args, model, device, train_loader, optimizer, epoch, ds_name)

        end = time.time()  # stop recording training time

        epoch_time = end - start

        # Evaluation on natural and adversarial examples
        print(
            "================================================================"
        )

        f.write("Epoch " + str(epoch) + "\n")

        print("Time for Training: " + str(epoch_time))

        f.write("Time for Training: " + str(epoch_time) + "\n")

        eval_clean(model, device, train_loader, "train", ds_name, f)

        eval_clean(model, device, test_loader, "test", ds_name, f)

        robust_eval(model, device, test_loader, ds_name, f)
        print(
            "================================================================"
        )

        # Save the model (if designated)

        if epoch % args.save_freq == 0:
            torch.save(
                model.state_dict(),
                os.path.join(
                    model_dir,
                    f"model-warmup-{ds_name}-{mod_name}-epoch{epoch}.pt",
                ),
            )

            torch.save(
                optimizer.state_dict(),
                os.path.join(
                    model_dir,
                    f"opt-warmup-{ds_name}-{mod_name}-epoch{epoch}.tar",
                ),
            )

    # Report full training time
    end_tot = time.time()

    total_time = end_tot - start_tot

    print("Total training time: " + str(total_time))

    f.write("Total training time: " + str(total_time) + "\n")

    f.close()  # close output file


if __name__ == "__main__":
    main_warmup("cifar100", "wideres34", 1)

    main_warmup("cifar10", "wideres34", 1)



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

import torch
import torch.nn.functional as F
from autoattack import AutoAttack
from torch import nn, optim
from torch.autograd import Variable

from airtk.attack import create_attack
from airtk.helper_functions import (
    cw_whitebox,
    mim_whitebox,
)


def adt_va_loss(model, x_natural, y, att, ds_name, optimizer, device):
    """

    Loss function for the ADT++ method. For each batch, the algorithm optimizes a distribution

    of adversarial examples surrounding calculated adversarial images and calculates the associated
    loss to the model.


    :param model: trained image classifier

    :param x_natural: batch of clean images from the training set

    :param y: set of correct labels for the clean images

    :param optimizer: optimizer for gradient calculation

    :param learning_rate: learning rate for the optimizer for distribution optimization

    :param epsilon: epsilon value for calculating perturbations

    :param perturb_steps: number of steps to use to calculate the distribution

    :param num_samples: number of samples to draw for distribution calculation

    :param lbd: lambda entropy multiplier


    :return cross-entropy loss between adversarial example drawn from distribution and correct labels

    """

    model.eval()  # put the model in evaluation mode

    # Perturb the example using VA

    x_adv = va_get_xadv(model, x_natural, y, att, device, ds_name).to(device)

    # Calculate the ADT loss

    loss = adt_loss(model, x_adv, y, optimizer)

    return loss


def clean_loss(model, x, y, optimizer):
    """

    Standard training loss function with no adversarial attack.


    :param model: image classifier

    :param x: batch of clean images

    :param y: correct labels for the clean images in x

    :param optimizer: optimizer for the model


    :return loss between correct labels and model output for the batch x

    """

    optimizer.zero_grad()  # zero out the gradients

    # Calculate cross-entropy loss between model output and correct labels

    out = model(x)

    criterion = nn.CrossEntropyLoss()

    loss = criterion(out, y)

    # Format predictions and batch metrics for return

    preds = out.detach()

    batch_metrics = {"loss": loss.item(), "clean_acc": accuracy(y, preds)}

    return loss, batch_metrics


def va_get_xadv(model, x, y, att, device, ds):
    """

    Adversarial training loss function for later various atts epoch, which simply

    calculates the loss based on the given attack.


    :param model: trained image classifier

    :param x: batch of clean images

    :param y: labels associated with images in x

    :param optimizer: optimizer to be updated

    :param att: attack to be used

    :param device: current device

    :param keep_clean: True if clean examples are kept


    :return loss to the classifier

    """

    # Set the criterion to cross-entropy loss

    criterion = nn.CrossEntropyLoss()

    # Create the adversary for AutoAttack

    adversary = AutoAttack(
        model, norm="Linf", eps=0.031, version="custom", verbose=False
    )

    # Based on the attack type, calculates the adversarial example

    if att == "linf-pgd-40":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 40, 0.01)

        x_adv, _ = attack.perturb(x, y)

    elif att == "cw":
        x_adv = cw_whitebox(model, x, y, device, ds)

    elif att == "mim":
        x_adv = mim_whitebox(model, x, y, device)

    elif att == "linf-pgd-3":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 3, 0.01)

        x_adv, _ = attack.perturb(x, y)

    elif att == "linf-pgd-20":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 20, 0.01)

        x_adv, _ = attack.perturb(x, y)

    elif att == "linf-pgd-7":
        attack = create_attack(model, criterion, "linf-pgd", 0.03, 7, 0.01)

        x_adv, _ = attack.perturb(x, y)

    elif att == "apgd-ce":
        adversary.attacks_to_run = ["apgd-ce"]

        x_adv = adversary.run_standard_evaluation(x, y)

    elif att == "apgd-dlr":
        adversary.attacks_to_run = ["apgd-dlr"]

        x_adv = adversary.run_standard_evaluation(x, y)

    elif att == "apgd-t":
        adversary.attacks_to_run = ["apgd-t"]

        x_adv = adversary.run_standard_evaluation(x, y)

    elif att == "fab":
        adversary.attacks_to_run = ["fab"]

        x_adv = adversary.run_standard_evaluation(x, y)

    elif att == "fab-t":
        adversary.attacks_to_run = ["fab-t"]

        x_adv = adversary.run_standard_evaluation(x, y)

    elif att == "square":
        adversary.attacks_to_run = ["square"]

        x_adv = adversary.run_standard_evaluation(x, y)

    else:
        raise NotImplementedError

    return x_adv


def adt_loss(
    model,
    x_natural,
    y,
    optimizer,
    learning_rate=1.0,
    epsilon=8.0 / 255.0,
    perturb_steps=10,
    num_samples=10,
    lbd=0.0,
):
    """

    Loss function for the Adversarial Distributional Training method. For each batch, the algorithm

    optimizes a distribution of adversarial examples surrounding the images and calculates the associated
    loss to the model.


    :param model: trained image classifier

    :param x_natural: batch of clean images from the training set

    :param y: set of correct labels for the clean images

    :param optimizer: optimizer for gradient calculation

    :param learning_rate: learning rate for the optimizer for distribution optimization

    :param epsilon: epsilon value for calculating perturbations

    :param perturb_steps: number of steps to use to calculate the distribution

    :param num_samples: number of samples to draw for distribution calculation

    :param lbd: lambda entropy multiplier


    :return cross-entropy loss between adversarial example drawn from distribution and correct labels

    """

    model.eval()  # put the model in evaluation mode

    # set the mean to a set of zeros the same size as the unperturbed image

    mean = Variable(torch.zeros(x_natural.size()).cuda(), requires_grad=True)

    # set the var to a set of zeros the same size as the unperturbed image

    var = Variable(torch.zeros(x_natural.size()).cuda(), requires_grad=True)

    # set the optimizer as an Adam optimizer based on the mean and var as model parameters

    optimizer_adv = optim.Adam([mean, var], lr=learning_rate, betas=(0.0, 0.0))

    # For each perturbation step

    for _ in range(
        perturb_steps
    ):  # perturb steps = number of steps distribution optimizer
        # For each sample to be drawn

        for s in range(
            num_samples
        ):  # num_samples = number of samples drawn from distribution before optimizing
            # set adv_std to the softplus (smooth + function) of the variation

            adv_std = F.softplus(var)

            # set rand-noise to a tensor the same size as x_natural filled randomly with mean 0 and var 1

            rand_noise = torch.randn_like(x_natural)

            # set adv to the hyperbolic tan of the mean plus the random noise times standard deviation

            adv = torch.tanh(
                mean + rand_noise * adv_std
            )  # Monte-Carlo sampling from distribution

            # omit the constants in -logp

            # set negative_logp to the following complex equation, representing the entropy

            negative_logp = (
                (rand_noise**2) / 2.0
                + (adv_std + 1e-8).log()
                + (1 - adv**2 + 1e-8).log()
            )

            entropy = (
                negative_logp.mean()
            )  # entropy is the average of negative_logp

            # set x_adv to the x_natural + epsilon times adv, clamped between 0 and 1

            x_adv = torch.clamp(x_natural + epsilon * adv, 0.0, 1.0)

            # minimize the negative loss

            # calculate the loss as the cross entropy between the classification of x_adv and y

            # minus the entropy times lambda, then step backwards

            with torch.enable_grad():
                loss = -F.cross_entropy(model(x_adv), y) - lbd * entropy

            loss.backward(retain_graph=s != num_samples - 1)

        # Step the distribution optimizer

        optimizer_adv.step()

    # set x_adv to x_natural plus epsilon times the hyperbolic tangent of the mean plus the softplus of the var

    # times random numbers like x_natual, clamped between 0 and 1

    x_adv = torch.clamp(
        x_natural
        + epsilon
        * torch.tanh(mean + F.softplus(var) * torch.randn_like(x_natural)),
        0.0,
        1.0,
    )

    model.train()  # set the model to training mode

    # set x_adv to a variable version of x_adv clamped between 0 and 1

    x_adv = Variable(torch.clamp(x_adv, 0.0, 1.0), requires_grad=False)

    optimizer.zero_grad()  # zero the gradient

    # Calculate the robust loss

    logits = model(x_adv)

    loss = F.cross_entropy(logits, y)
    return loss


def standard_loss(model, x, y, optimizer):
    """

    Standard training loss function with no adversarial attack.


    :param model: image classifier

    :param x: batch of clean images

    :param y: correct labels for the clean images in x

    :param optimizer: optimizer for the model


    :return loss between correct labels and model output for the batch x

    """

    optimizer.zero_grad()  # zero out the gradients

    # Calculate cross-entropy loss between model output and correct labels

    out = model(x)

    criterion = nn.CrossEntropyLoss()

    loss = criterion(out, y)

    # Format predictions and batch metrics for return

    preds = out.detach()

    batch_metrics = {"loss": loss.item(), "clean_acc": accuracy(y, preds)}

    return loss, batch_metrics


def standardat_loss(model, x, y, optimizer, keep_clean=False):
    """

    Basic loss function from Madry et al. calculating an adversarial sample and the

    loss resulting.


    :param model: trained image classifier

    :param x: batch of clean of images

    :param y: labels for images in x

    :param optimizer: optimizer associated with the model

    :param attack: attack used to perturb clean images

    :param keep_clean: True if clean examples are kept


    :return cross-entropy loss between the adversarial output and correct class

    """

    # Calculate the adversarial example based on the atts

    criterion = nn.CrossEntropyLoss()

    attack = create_attack(model, criterion, "linf-pgd", 0.1, 40, 0.01)

    x_adv, _ = attack.perturb(x, y)

    optimizer.zero_grad()  # zero out the gradients

    # Correctly format x_adv and y_adv

    if keep_clean:
        x_adv = torch.cat((x, x_adv), dim=0)

        y_adv = torch.cat((y, y), dim=0)

    else:
        y_adv = y

    # Use adversarial output to calculate the cross-entropy loss with correct labels

    out = model(x_adv)

    criterion = nn.CrossEntropyLoss()

    loss = criterion(out, y_adv)

    return loss


def accuracy(true, preds):
    """

    Computes multi-class accuracy.


    :param true: true labels

    :param preds: predicted labels

    :return multi-class accuracy

    """

    accuracy = (
        torch.softmax(preds, dim=1).argmax(dim=1) == true
    ).sum().float() / float(true.size(0))

    return accuracy.item()


def CWLoss(output, target, confidence=0):
    """

    CW loss (Marging loss).

    """

    num_classes = output.shape[-1]

    target = target.data

    target_onehot = torch.zeros(target.size() + (num_classes,))

    target_onehot = target_onehot.cuda()

    target_onehot.scatter_(1, target.unsqueeze(1), 1.0)

    target_var = Variable(target_onehot, requires_grad=False)

    real = (target_var * output).sum(1)

    other = ((1.0 - target_var) * output - target_var * 10000.0).max(1)[0]

    loss = -torch.clamp(real - other + confidence, min=0.0)

    loss = torch.sum(loss)
    return loss



from ..trainer import Trainer
from ...utils import get_subloader
from collections import OrderedDict

r"""
Base class for adversarial trainers.

Functions:
    self.record_rob : function for recording standard accuracy and robust accuracy against FGSM, PGD, and GN.

"""


class AdvTrainer(Trainer):
    def __init__(self, rmodel, device=None):
        super().__init__(rmodel, device)
        self.rob_dict = OrderedDict()

    def record_rob(
        self,
        train_loader,
        val_loader,
        eps=None,
        alpha=None,
        steps=None,
        std=None,
        n_train_limit=None,
        n_val_limit=None,
    ):
        if (alpha is None) and (steps is not None):
            raise ValueError("Both alpha and steps should be given for PGD.")
        elif (alpha is not None) and (steps is None):
            raise ValueError("Both alpha and steps should be given for PGD.")

        self.rob_dict["train_loader"] = get_subloader(train_loader, n_train_limit)
        self.rob_dict["val_loader"] = get_subloader(val_loader, n_val_limit)
        self.rob_dict["loaders"] = {
            "(Tr)": self.rob_dict["train_loader"],
            "(Val)": self.rob_dict["val_loader"],
        }
        self.rob_dict["eps"] = eps
        self.rob_dict["alpha"] = alpha
        self.rob_dict["steps"] = steps
        self.rob_dict["std"] = std

    def record_during_eval(self):
        for flag, loader in self.rob_dict["loaders"].items():
            self.dict_record["Clean" + flag] = self.rmodel.eval_accuracy(loader)

            eps = self.rob_dict.get("eps")
            if eps is not None:
                self.dict_record["FGSM" + flag] = self.rmodel.eval_rob_accuracy_fgsm(
                    loader, eps=eps, verbose=False
                )
                steps = self.rob_dict.get("steps")
                alpha = self.rob_dict.get("alpha")
                if steps is not None:
                    self.dict_record["PGD" + flag] = self.rmodel.eval_rob_accuracy_pgd(
                        loader, eps=eps, alpha=alpha, steps=steps, verbose=False
                    )

            std = self.rob_dict.get("std")
            if std is not None:
                self.dict_record["GN" + flag] = self.rmodel.eval_rob_accuracy_gn(
                    loader, std=std, verbose=False
                )



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
from collections import OrderedDict

import torch
import torch.nn.functional as F

EPS = 1e-20


def diff_in_weights(model, proxy):
    diff_dict = OrderedDict()
    model_state_dict = model.state_dict()
    proxy_state_dict = proxy.state_dict()
    for (old_k, old_w), (new_k, new_w) in zip(
        model_state_dict.items(), proxy_state_dict.items()
    ):
        if len(old_w.size()) <= 1:
            continue
        if "weight" in old_k:
            diff_w = new_w - old_w
            diff_dict[old_k] = old_w.norm() / (diff_w.norm() + EPS) * diff_w
    return diff_dict


def add_into_weights(model, diff, coeff=1.0):
    names_in_diff = diff.keys()
    with torch.no_grad():
        for name, param in model.named_parameters():
            if name in names_in_diff:
                param.add_(coeff * diff[name])


class TradesAWP:
    def __init__(self, model, proxy, proxy_optim, gamma):
        super(TradesAWP, self).__init__()
        self.model = model
        self.proxy = proxy
        self.proxy_optim = proxy_optim
        self.gamma = gamma

    def calc_awp(self, inputs_adv, inputs_clean, targets, beta):
        self.proxy.load_state_dict(self.model.state_dict())
        self.proxy.train()

        loss_natural = F.cross_entropy(self.proxy(inputs_clean), targets)
        loss_robust = F.kl_div(
            F.log_softmax(self.proxy(inputs_adv), dim=1),
            F.softmax(self.proxy(inputs_clean), dim=1),
            reduction="batchmean",
        )
        loss = -1.0 * (loss_natural + beta * loss_robust)

        self.proxy_optim.zero_grad()
        loss.backward()
        self.proxy_optim.step()

        # the adversary weight perturb
        diff = diff_in_weights(self.model, self.proxy)
        return diff

    def perturb(self, diff):
        add_into_weights(self.model, diff, coeff=1.0 * self.gamma)

    def restore(self, diff):
        add_into_weights(self.model, diff, coeff=-1.0 * self.gamma)



import warnings

import torch

from .. import RobModel
from ..utils import load_model
from ._parse_notion import get_link_by_id


def load_pretrained(id, flag, save_dir="./"):
    assert flag in ['Init', 'Best', 'Last']
    url = get_link_by_id(id, flag)
    state_dict = torch.hub.load_state_dict_from_url(url,
                                                    progress=True, model_dir=save_dir,
                                                    file_name="%s_%s.pth"%(id, flag),
                                                    map_location='cpu')
    return _construct_rmodel_from_dict(id, state_dict)


def _construct_rmodel_from_dict(id, state_dict):
    model_name = id.split("_")[1]
    n_classes = state_dict['rmodel']['n_classes'].item()
    model = load_model(model_name=model_name, n_classes=n_classes)
    rmodel = RobModel(model, n_classes=n_classes)
    rmodel.load_state_dict_auto(state_dict['rmodel'])
    return rmodel



import copy
import itertools
import os
from random import sample, shuffle

import torch
import torch.nn as nn

from ..attack import Attack
from ..attacks.bim import BIM

# fail-safe import of tqdm
try:
    from tqdm import tqdm
except ImportError:

    def tqdm(iterator, *args, **kwargs):
        return iterator


class LGV(Attack):
    r"""
    LGV attack in the paper 'LGV: Boosting Adversarial Example Transferability from Large Geometric Vicinity'
    [https://arxiv.org/abs/2207.13129]

    Arguments:
        model (nn.Module): initial model to attack.
        trainloader (torch.utils.data.DataLoader): data loader of the unnormalized train set. Must load data in [0, 1].
        Be aware that the batch size may impact success rate. The original paper uses a batch size of 256. A different
        batch-size might require to tune the learning rate.
        lr (float): constant learning rate to collect models. In the paper, 0.05 is best for ResNet-50. 0.1 seems best
        for some other architectures. (Default: 0.05)
        epochs (int): number of epochs. (Default: 10)
        nb_models_epoch (int): number of models to collect per epoch. (Default: 4)
        wd (float): weight decay of SGD to collect models. (Default: 1e-4)
        n_grad (int): number of models to ensemble at each attack iteration. 1 (default) is recommended for efficient
        iterative attacks. Higher numbers give generally better results at the expense of computations. -1 uses all
        models (should be used for single-step attacks like FGSM).
        verbose (bool): print progress. Install the tqdm package for better print. (Default: True)

    .. note:: If a list of models is not provided to `load_models()`, the attack will start by collecting models along
    the SGD trajectory for `epochs` epochs with the constant learning rate `lr`.

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`, `H = height`
        and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.LGV(model, trainloader, lr=0.05, epochs=10, nb_models_epoch=4, wd=1e-4, n_grad=1, attack_class=BIM, eps=4/255, alpha=4/255/10, steps=50, verbose=True)
        >>> attack.collect_models()
        >>> attack.save_models('./models/lgv/')
        >>> adv_images = attack(images, labels)
    """

    def __init__(
        self,
        model,
        trainloader,
        lr=0.05,
        epochs=10,
        nb_models_epoch=4,
        wd=1e-4,
        n_grad=1,
        verbose=True,
        attack_class=BIM,
        **kwargs,
    ):
        model = copy.deepcopy(model)  # deep copy the model to train it
        super().__init__("LGV", model)
        self.trainloader = trainloader
        self.lr = lr
        self.epochs = epochs
        self.nb_models_epoch = nb_models_epoch
        self.wd = wd
        self.n_grad = n_grad
        self.order = "shuffle"
        self.attack_class = attack_class
        self.verbose = verbose
        self.kwargs_att = kwargs
        if not isinstance(lr, float) or lr < 0:
            raise ValueError("lr should be a non-negative float")
        if not isinstance(epochs, int) or epochs < 0:
            raise ValueError("epochs should be a non-negative integer")
        if not isinstance(nb_models_epoch, int) or nb_models_epoch < 0:
            raise ValueError(
                "nb_models_epoch should be a non-negative integer"
            )
        self.supported_mode = ["default", "targeted"]
        self.list_models = []
        self.base_attack = None  # will be initialized after model collection

    def collect_models(self):
        """
        Collect LGV models along the SGD trajectory
        """
        given_training = self.model.training
        self.model.train()
        optimizer = torch.optim.SGD(
            self.model.parameters(),
            lr=self.lr,
            momentum=0.9,
            weight_decay=self.wd,
        )
        loss_fn = nn.CrossEntropyLoss()
        epoch_frac = 1.0 / self.nb_models_epoch
        n_batches = int(len(self.trainloader) * epoch_frac)
        for i_sample in tqdm(
            range(self.epochs * self.nb_models_epoch), "Collecting models"
        ):
            loader = itertools.islice(self.trainloader, n_batches)
            for j, (input, target) in enumerate(loader):
                if torch.cuda.is_available():
                    input = input.to("cuda", non_blocking=True)
                    target = target.to("cuda", non_blocking=True)
                pred = self.get_logits(input)
                loss = loss_fn(pred, target)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
            model_sample = copy.deepcopy(self.model)
            if not given_training:
                model_sample.eval()
            self.list_models.append(model_sample)
        if not given_training:
            self.model.eval()

    def load_models(self, list_models):
        """
        Load collected models

        Arguments:
        list_models (list of nn.Module): list of LGV models.
        """
        if not isinstance(list_models, list):
            raise ValueError("list_models should be a list of pytorch models")
        self.list_models = list_models

    def save_models(self, path):
        """
        Save collected models to the `path` directory

        Arguments:
        path (str): directory where to save models.
        """
        if len(self.list_models) == 0:
            raise RuntimeError(
                "Call collect_models() before saving collected models."
            )
        os.makedirs(path, exist_ok=True)
        for i, model in enumerate(self.list_models):
            path_i = os.path.join(path, f"lgv_model_{i:05}.pt")
            torch.save({"state_dict": model.state_dict()}, path_i)

    def forward(self, images, labels):
        r"""
        Overridden.
        """
        if len(self.list_models) == 0:
            if self.verbose:
                print(f"Phase 1: collect models for {self.epochs} epochs")
            self.collect_models()

        if not self.base_attack:
            if self.verbose:
                print(
                    f"Phase 2: craft adversarial examples with {self.attack_class.__name__}"
                )
            self.list_models = [
                model.to(self.device) for model in self.list_models
            ]
            f_model = LightEnsemble(
                self.list_models, order=self.order, n_grad=self.n_grad
            )
            if self._model_training:
                f_model.eval()
            self.base_attack = self.attack_class(
                model=f_model.to(self.device), **self.kwargs_att
            )
        # set_model_training_mode() to base attack
        self.base_attack.set_model_training_mode(
            model_training=self._model_training,
            batchnorm_training=self._batchnorm_training,
            dropout_training=self._dropout_training,
        )
        # set targeted to base attack
        if self.targeted:
            if self.attack_mode == "targeted":
                self.base_attack.set_mode_targeted_by_function(
                    target_map_function=self._target_map_function
                )
            elif self.attack_mode == "targeted(least-likely)":
                self.base_attack.set_mode_targeted_least_likely(
                    kth_min=self._kth_min
                )
            elif self.attack_mode == "targeted(random)":
                self.base_attack.set_mode_targeted_random()
            else:
                raise NotImplementedError(
                    "Targeted attack mode not supported by LGV."
                )
        # set return type to base attack
        # self.base_attack.set_return_type(self.return_type)

        adv_images = self.base_attack(images, labels)
        return adv_images


class LightEnsemble(nn.Module):
    def __init__(self, list_models, order="shuffle", n_grad=1):
        """
        Perform a single forward pass to one of the models when call forward()

        Arguments:
            list_models (list of nn.Module): list of LGV models.
            order (str): 'shuffle' draw a model without replacement (default), 'random' draw a model with replacement,
            None cycle in provided order.
            n_grad (int): number of models to ensemble in each forward pass (fused logits). Select models according to
            `order`. If equal to -1, use all models and order is ignored.
        """
        super(LightEnsemble, self).__init__()
        self.n_models = len(list_models)
        if self.n_models < 1:
            raise ValueError("Empty list of models")
        if not (n_grad > 0 or n_grad == -1):
            raise ValueError(
                "n_grad should be strictly positive or equal to -1"
            )
        if order == "shuffle":
            shuffle(list_models)
        elif order in [None, "random"]:
            pass
        else:
            raise ValueError("Not supported order")
        self.models = nn.ModuleList(list_models)
        self.order = order
        self.n_grad = n_grad
        self.f_count = 0

    def forward(self, x):
        if self.n_grad >= self.n_models or self.n_grad < 0:
            indexes = list(range(self.n_models))
        elif self.order == "random":
            indexes = sample(range(self.n_models), self.n_grad)
        else:
            indexes = [
                i % self.n_models
                for i in list(range(self.f_count, self.f_count + self.n_grad))
            ]
            self.f_count += self.n_grad
        if self.n_grad == 1:
            x = self.models[indexes[0]](x)
        else:
            # clone to make sure x is not changed by inplace methods
            x_list = [
                model(x.clone())
                for i, model in enumerate(self.models)
                if i in indexes
            ]
            x = torch.stack(x_list)
            x = torch.mean(x, dim=0, keepdim=False)
        return x



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).

import copy
import os
import time
from collections import namedtuple

import torch
from torch import optim

from airtk.defense.trades.trades_loss import trades_loss
from airtk.helper_functions import (
    adjust_learning_rate,
    eval_clean,
    load_data,
    robust_eval,
)
from airtk.model import (
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
    WideResNet,
)


class TradesTraining:
    def __init__(
        self,
        dataset_name: str,
        model_name: str,
        batch_size: int = 128,
        epochs: int = 100,
        weight_decay: float = 2e-4,
        lr: float = 0.1,
        momentum: float = 0.9,
        seed: int = 1,
        log_interval: int = 1,
        model_dir: str = "./data/model",
        save_freq: int = 1,
        lr_schedule: str = "decay",
    ) -> None:
        ArgsPrototype = namedtuple(
            "ArgsPrototype",
            [
                "batch_size",
                "test_batch_size",
                "epochs",
                "weight_decay",
                "lr",
                "momentum",
                "no_cuda",
                "seed",
                "log_interval",
                "model_dir",
                "save_freq",
                "lr_schedule",
            ],
        )
        self._args = ArgsPrototype(
            batch_size,
            batch_size,
            epochs,
            weight_decay,
            lr,
            momentum,
            False,
            seed,
            log_interval,
            model_dir,
            save_freq,
            lr_schedule,
        )
        self._model_name = model_name
        self._dataset_name = dataset_name

    def __call__(self) -> None:
        torch.manual_seed(self._args.seed)
        if not os.path.exists(self._args.model_dir):
            os.makedirs(self._args.model_dir)

        main_trades(self._dataset_name, self._model_name, self._args)


# parser = argparse.ArgumentParser(
# description="PyTorch CIFAR Adversarial Training Framework"
# )
# parser.add_argument(
# "--batch-size",
# type=int,
# default=128,
# metavar="N",
# help="input batch size for training (default: 128)",
# )
# parser.add_argument(
# "--test-batch-size",
# type=int,
# default=128,
# metavar="N",
# help="input batch size for testing (default: 128)",
# )
# parser.add_argument(
# "--epochs",
# type=int,
# default=100,
# metavar="N",
# help="number of epochs to train",
# )
# parser.add_argument(
# "--weight-decay", "--wd", default=2e-4, type=float, metavar="W"
# )
# parser.add_argument(
# "--lr", type=float, default=0.1, metavar="LR", help="learning rate"
# )
# parser.add_argument(
# "--momentum", type=float, default=0.9, metavar="M", help="SGD momentum"
# )
# parser.add_argument(
# "--no-cuda",
# action="store_true",
# default=False,
# help="disables CUDA training",
# )
# parser.add_argument(
# "--seed", type=int, default=1, metavar="S", help="random seed (default: 1)"
# )
# parser.add_argument(
# "--log-interval",
# type=int,
# default=100,
# metavar="N",
# help="how many batches to wait before logging training status",
# )
# parser.add_argument(
# "--model-dir",
# default="./data/model",
# help="directory of model for saving checkpoint",
# )
# parser.add_argument(
# "--save-freq",
# "-s",
# default=1,
# type=int,
# metavar="N",
# help="save frequency",
# )
# parser.add_argument(
# "--lr-schedule",
# default="decay",
# help="schedule for adjusting learning rate",
# )
# args = parser.parse_args()

# Establish the settings
# model_dir = args.model_dir
# if not os.path.exists(model_dir):
# os.makedirs(model_dir)
# use_cuda = not args.no_cuda and torch.cuda.is_available()
# torch.manual_seed(args.seed)
# device = torch.device("cuda" if use_cuda else "cpu")
# kwargs = {"num_workers": 1, "pin_memory": True} if use_cuda else {}


def train(args, model, device, train_loader, optimizer, ds_name, epoch, f):
    """
    Trains one epoch by calculating the loss for each batch and updating the model

    :param args: set of arguments including learning rate, log interval, etc
    :param model: model to train
    :param device: current device
    :param train_loader: data loader containing the training dataset
    :param optimizer: optimizer to train
    :param epoch: current epoch of training
    """

    model.train()  # set the  model to training mode
    for batch_idx, sample in enumerate(train_loader):
        # Set the data and target (x and y) to the device
        if ds_name == "cifar100":
            data, target, coarse = sample
        else:
            data, target = sample
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()  # zero out the gradients

        # Calculate the clean loss
        loss = trades_loss(model, data, target, optimizer)

        # Update the model based on the loss
        loss.backward()
        optimizer.step()

        # print progress
        if batch_idx % args.log_interval == 0:
            print(
                f"Train Epoch: {epoch} [{batch_idx * len(data)}/{len(train_loader.dataset)} ({100.0 * batch_idx / len(train_loader):.0f}%)]\tLoss: {loss.item():.6f}"
            )
            f.write(
                f"Train Epoch: {epoch} [{batch_idx * len(data)}/{len(train_loader.dataset)} ({100.0 * batch_idx / len(train_loader):.0f}%)]\tLoss: {loss.item():.6f}"
            )


def main_trades(ds_name, mod_name, args):
    """
    Main training method, which establishes the model/dataset before conducting training and
    clean testing for each epoch. Then reports final training time and conducts robustness tests
    on the final model.

    :param ds_name: the dataset to use for training
    :param mod_name: the model to use for training
    """

    # Set up file for printing the output
    filename = f"log/trades-{ds_name}-{mod_name}-output.txt"
    f = open(filename, "a")

    device = "cuda" if torch.cuda.is_available() else "cpu"
    # Initialize the desired model
    start_tot = time.time()  # start timing the full method
    if mod_name == "res18":
        model = ResNet18(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )
    elif mod_name == "res34":
        model = ResNet34(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )
    elif mod_name == "res50":
        model = ResNet50(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )
    elif mod_name == "res101":
        model = ResNet101(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )
    elif mod_name == "res152":
        model = ResNet152(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )
    elif mod_name == "wideres34":
        model = WideResNet(
            depth=34, num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)
    else:
        raise NotImplementedError

    # Set up the optimizer to the arguments
    optimizer = optim.SGD(
        model.parameters(),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    # Set up the dataloaders
    kwargs = {"num_workers": 1, "pin_memory": True} if device == "cuda" else {}
    train_loader, test_loader = load_data(ds_name, args, kwargs, coarse=True)

    # Save the model and optimizer
    model_dir = args.model_dir
    torch.save(
        model.state_dict(),
        os.path.join(model_dir, f"model-trades-{ds_name}-{mod_name}-start.pt"),
    )
    torch.save(
        optimizer.state_dict(),
        os.path.join(model_dir, f"opt-trades-{ds_name}-{mod_name}-start.tar"),
    )

    # Begin training for the designated number of epochs
    for epoch in range(1, args.epochs + 1):
        # Adjust the learning rate for the SGD
        adjust_learning_rate(optimizer, epoch, args)

        # Commence training, timing the method and recording the time
        start = time.time()  # start timing the method
        train(args, model, device, train_loader, optimizer, ds_name, epoch, f)
        end = time.time()  # finish timing the method
        epoch_time = end - start

        # evaluation on natural examples (training and testing)
        print(
            "================================================================"
        )
        f.write("Epoch " + str(epoch) + "\n")
        print("Time for Training: " + str(epoch_time))
        f.write("Time for Training: " + str(epoch_time) + "\n")
        eval_clean(model, device, train_loader, "train", ds_name, f)
        eval_clean(model, device, test_loader, "test", ds_name, f)
        robust_eval(model, device, test_loader, ds_name, f)
        print(
            "================================================================"
        )

        # Save the model (if designated)
        if epoch % args.save_freq == 0:
            torch.save(
                model.state_dict(),
                os.path.join(
                    model_dir,
                    f"model-trades-{ds_name}-{mod_name}-epoch{epoch}.pt",
                ),
            )
            torch.save(
                optimizer.state_dict(),
                os.path.join(
                    model_dir,
                    f"opt-trades-{ds_name}-{mod_name}-epoch{epoch}.tar",
                ),
            )

    # Report the time taken to fully train the model
    end_tot = time.time()
    total_time = end_tot - start_tot
    print("Total training time: " + str(total_time))
    f.write("Total training time: " + str(total_time) + "\n")

    # Load the final trained model, make a deepcopy, and evaluate against FGSM and PGD
    if mod_name == "res18":
        trained_model = ResNet18(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)
    elif mod_name == "res34":
        trained_model = ResNet34(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)
    elif mod_name == "res50":
        trained_model = ResNet50(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)
    elif mod_name == "res101":
        trained_model = ResNet101(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)
    elif mod_name == "res152":
        trained_model = ResNet152(
            num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)
    elif mod_name == "wideres34":
        trained_model = WideResNet(
            depth=34, num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)
    else:
        raise NotImplementedError

    trained_model.load_state_dict(
        torch.load(
            f"data/model/model-trades-{ds_name}-{mod_name}-epoch{args.epochs}.pt"
        )
    )
    model_copy = copy.deepcopy(trained_model)
    robust_eval(model_copy, device, test_loader, ds_name, f)

    f.close()  # close the output file


# if __name__ == "__main__":
# main_trades("cifar100", "res18")



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

import numpy as np
import torch
from torch.distributions import laplace, uniform

from airtk.model import (
    ResNet18,
    ResNet34,
    ResNet50,
    ResNet101,
    ResNet152,
    WideResNet,
)


def replicate_input(x):
    """

    Clone the input tensor x.
    """

    return x.detach().clone()


def replicate_input_withgrad(x):
    """

    Clone the input tensor x and set requires_grad=True.
    """

    return x.detach().clone().requires_grad_()


def calc_l2distsq(x, y):
    """

    Calculate L2 distance between tensors x and y.
    """

    d = (x - y) ** 2

    return d.view(d.shape[0], -1).sum(dim=1)


def clamp(input, min=None, max=None):
    """

    Clamp a tensor by its minimun and maximun values.
    """
    ndim = input.ndimension()

    if min is None:
        pass

    elif isinstance(min, (float, int)):
        input = torch.clamp(input, min=min)

    elif isinstance(min, torch.Tensor):
        if min.ndimension() == ndim - 1 and min.shape == input.shape[1:]:
            input = torch.max(input, min.view(1, *min.shape))

        else:
            assert min.shape == input.shape

            input = torch.max(input, min)

    else:
        raise ValueError("min can only be None | float | torch.Tensor")

    if max is None:
        pass

    elif isinstance(max, (float, int)):
        input = torch.clamp(input, max=max)

    elif isinstance(max, torch.Tensor):
        if max.ndimension() == ndim - 1 and max.shape == input.shape[1:]:
            input = torch.min(input, max.view(1, *max.shape))

        else:
            assert max.shape == input.shape

            input = torch.min(input, max)

    else:
        raise ValueError("max can only be None | float | torch.Tensor")
    return input


def _batch_multiply_tensor_by_vector(vector, batch_tensor):
    """Equivalent to the following.

    for ii in range(len(vector)):

        batch_tensor.data[ii] *= vector[ii]

    return batch_tensor
    """
    return (
        (batch_tensor.transpose(0, -1) * vector).transpose(0, -1).contiguous()
    )


def _batch_clamp_tensor_by_vector(vector, batch_tensor):
    """Equivalent to the following.

    for ii in range(len(vector)):

        batch_tensor[ii] = clamp(

            batch_tensor[ii], -vector[ii], vector[ii])
    """
    return (
        torch.min(torch.max(batch_tensor.transpose(0, -1), -vector), vector)
        .transpose(0, -1)
        .contiguous()
    )


def batch_multiply(float_or_vector, tensor):
    """

    Multpliy a batch of tensors with a float or vector.
    """

    if isinstance(float_or_vector, torch.Tensor):
        assert len(float_or_vector) == len(tensor)

        tensor = _batch_multiply_tensor_by_vector(float_or_vector, tensor)

    elif isinstance(float_or_vector, float):
        tensor *= float_or_vector

    else:
        raise TypeError("Value has to be float or torch.Tensor")
    return tensor


def batch_clamp(float_or_vector, tensor):
    """

    Clamp a batch of tensors.
    """

    if isinstance(float_or_vector, torch.Tensor):
        assert len(float_or_vector) == len(tensor)

        tensor = _batch_clamp_tensor_by_vector(float_or_vector, tensor)
        return tensor

    elif isinstance(float_or_vector, float):
        tensor = clamp(tensor, -float_or_vector, float_or_vector)

    else:
        raise TypeError("Value has to be float or torch.Tensor")
    return tensor


def _get_norm_batch(x, p):
    """

    Returns the Lp norm of batch x.
    """

    batch_size = x.size(0)

    return x.abs().pow(p).view(batch_size, -1).sum(dim=1).pow(1.0 / p)


def _thresh_by_magnitude(theta, x):
    """

    Threshold by magnitude.
    """

    return torch.relu(torch.abs(x) - theta) * x.sign()


def clamp_by_pnorm(x, p, r):
    """

    Clamp tensor by its norm.
    """
    assert isinstance(p, float) or isinstance(p, int)

    norm = _get_norm_batch(x, p)

    if isinstance(r, torch.Tensor):
        assert norm.size() == r.size()

    else:
        assert isinstance(r, float)

    factor = torch.min(r / norm, torch.ones_like(norm))

    return batch_multiply(factor, x)


def is_float_or_torch_tensor(x):
    """

    Return whether input x is a float or a torch.Tensor.
    """

    return isinstance(x, torch.Tensor) or isinstance(x, float)


def normalize_by_pnorm(x, p=2, small_constant=1e-6):
    """

    Normalize gradients for gradient (not gradient sign) atts.

    Arguments:

        x (torch.Tensor): tensor containing the gradients on the input.

        p (int): (optional) order of the norm for the normalization (1 or 2).

        small_constant (float): (optional) to avoid dividing by zero.

    Returns:

        normalized gradients.
    """
    assert isinstance(p, float) or isinstance(p, int)

    norm = _get_norm_batch(x, p)

    norm = torch.max(norm, torch.ones_like(norm) * small_constant)

    return batch_multiply(1.0 / norm, x)


def rand_init_delta(delta, x, ord, eps, clip_min, clip_max):
    """

    Randomly initialize the perturbation.
    """

    if isinstance(eps, torch.Tensor):
        assert len(eps) == len(delta)

    if ord == np.inf:
        delta.data.uniform_(-1, 1)

        delta.data = batch_multiply(eps, delta.data)

    elif ord == 2:
        delta.data.uniform_(clip_min, clip_max)

        delta.data = delta.data - x

        delta.data = clamp_by_pnorm(delta.data, ord, eps)

    elif ord == 1:
        ini = laplace.Laplace(
            loc=delta.new_tensor(0), scale=delta.new_tensor(1)
        )
        delta.data = ini.sample(delta.data.shape)

        delta.data = normalize_by_pnorm(delta.data, p=1)

        ray = uniform.Uniform(0, eps).sample()

        delta.data *= ray

        delta.data = clamp(x.data + delta.data, clip_min, clip_max) - x.data

    else:
        error = "Only ord = inf, ord = 1 and ord = 2 have been implemented"

        raise NotImplementedError(error)

    delta.data = clamp(x + delta.data, min=clip_min, max=clip_max) - x
    return delta.data


def get_model(mod_name, ds_name, device):
    if mod_name == "res18":
        model = ResNet18(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res34":
        model = ResNet34(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res50":
        model = ResNet50(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res101":
        model = ResNet101(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "res152":
        model = ResNet152(num_classes=100 if ds_name == "cifar100" else 10).to(
            device
        )

    elif mod_name == "wideres34":
        model = WideResNet(
            depth=34, num_classes=100 if ds_name == "cifar100" else 10
        ).to(device)

    else:
        raise NotImplementedError

    return model



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).
from airtk.attack.apgd import (
    L2APGDAttack,
    LinfAPGDAttack,
)
from airtk.attack.base import Attack
from airtk.attack.fgsm import (
    FGMAttack,
    FGSMAttack,
    L2FastGradientAttack,
    LinfFastGradientAttack,
)
from airtk.attack.pgd import (
    L2PGDAttack,
    LinfPGDAttack,
    PGDAttack,
)
from airtk.attack.util import create_attack

__all__ = [
    "Attack",
    "create_attack",
    "LinfAPGDAttack",
    "L2APGDAttack",
    "FGMAttack",
    "FGSMAttack",
    "L2FastGradientAttack",
    "LinfFastGradientAttack",
    "PGDAttack",
    "L2PGDAttack",
    "LinfPGDAttack",
]



import torch
import torchvision
import torchvision.transforms as transforms


def create_train_dataset(ds_name, batch_size=128, root="../data"):
    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            # transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ]
    )
    if ds_name == "cifar10":
        trainset = torchvision.datasets.CIFAR10(
            root=root, train=True, download=True, transform=transform_train
        )
    elif ds_name == "cifar100":
        trainset = torchvision.datasets.CIFAR100(
            root=root, train=True, download=True, transform=transform_train
        )
    # trainset = torchvision.datasets.MNIST(root=root, train=True, download=True, transform=transform_train)
    trainloader = torch.utils.data.DataLoader(
        trainset, batch_size=batch_size, shuffle=True, num_workers=2
    )

    return trainloader


def create_test_dataset(ds_name, batch_size=128, root="../data"):
    transform_test = transforms.Compose(
        [
            transforms.ToTensor(),
            # transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ]
    )
    if ds_name == "cifar10":
        testset = torchvision.datasets.CIFAR10(
            root=root, train=False, download=True, transform=transform_test
        )
    elif ds_name == "cifar100":
        testset = torchvision.datasets.CIFAR100(
            root=root, train=False, download=True, transform=transform_test
        )
    # testset = torchvision.datasets.MNIST(root=root, train=False, download=True, transform=transform_test)
    testloader = torch.utils.data.DataLoader(
        testset, batch_size=100, shuffle=False, num_workers=2
    )
    return testloader


if __name__ == "__main__":
    print(create_train_dataset("cifar10"))
    print(create_test_dataset("cifar10"))



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
import math

import torch
import torch.nn as nn
import torch.nn.functional as F


class BasicBlock(nn.Module):
    def __init__(self, in_planes, out_planes, stride, dropRate=0.0):
        super(BasicBlock, self).__init__()
        self.bn1 = nn.BatchNorm2d(in_planes)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(out_planes)
        self.relu2 = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(
            out_planes,
            out_planes,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
        )
        self.droprate = dropRate
        self.equalInOut = in_planes == out_planes
        self.convShortcut = (
            (not self.equalInOut)
            and nn.Conv2d(
                in_planes,
                out_planes,
                kernel_size=1,
                stride=stride,
                padding=0,
                bias=False,
            )
            or None
        )

    def forward(self, x):
        if not self.equalInOut:
            x = self.relu1(self.bn1(x))
        else:
            out = self.relu1(self.bn1(x))
        out = self.relu2(self.bn2(self.conv1(out if self.equalInOut else x)))
        if self.droprate > 0:
            out = F.dropout(out, p=self.droprate, training=self.training)
        out = self.conv2(out)
        return torch.add(x if self.equalInOut else self.convShortcut(x), out)


class NetworkBlock(nn.Module):
    def __init__(
        self, nb_layers, in_planes, out_planes, block, stride, dropRate=0.0
    ):
        super(NetworkBlock, self).__init__()
        self.layer = self._make_layer(
            block, in_planes, out_planes, nb_layers, stride, dropRate
        )

    def _make_layer(
        self, block, in_planes, out_planes, nb_layers, stride, dropRate
    ):
        layers = []
        for i in range(int(nb_layers)):
            layers.append(
                block(
                    i == 0 and in_planes or out_planes,
                    out_planes,
                    i == 0 and stride or 1,
                    dropRate,
                )
            )
        return nn.Sequential(*layers)

    def forward(self, x):
        return self.layer(x)


class WideResNet(nn.Module):
    def __init__(self, depth, num_classes, widen_factor=1, dropRate=0.0):
        super(WideResNet, self).__init__()
        nChannels = [
            16,
            16 * widen_factor,
            32 * widen_factor,
            64 * widen_factor,
        ]
        assert (depth - 4) % 6 == 0
        n = (depth - 4) / 6
        block = BasicBlock
        # 1st conv before any network block
        self.conv1 = nn.Conv2d(
            3, nChannels[0], kernel_size=3, stride=1, padding=1, bias=False
        )
        # 1st block
        self.block1 = NetworkBlock(
            n, nChannels[0], nChannels[1], block, 1, dropRate
        )
        # 2nd block
        self.block2 = NetworkBlock(
            n, nChannels[1], nChannels[2], block, 2, dropRate
        )
        # 3rd block
        self.block3 = NetworkBlock(
            n, nChannels[2], nChannels[3], block, 2, dropRate
        )
        # global average pooling and classifier
        self.bn1 = nn.BatchNorm2d(nChannels[3])
        self.relu = nn.ReLU(inplace=True)
        self.fc = nn.Linear(nChannels[3], num_classes)
        self.nChannels = nChannels[3]

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                m.bias.data.zero_()

    def forward(self, x):
        out = self.conv1(x)
        out = self.block1(out)
        out = self.block2(out)
        out = self.block3(out)
        out = self.relu(self.bn1(out))
        out = F.avg_pool2d(out, 8)
        out = out.view(-1, self.nChannels)

        return self.fc(out), out.view(x.size(0), -1)
